#ifndef IRRAM_REALMATRIX_H
#define IRRAM_REALMATRIX_H

#include "REALS.h" 

class REALMATRIX : public METRIC_OBJECT
{
public:
// Constructors: -------------------------------

REALMATRIX(unsigned int rows,unsigned int columns);
REALMATRIX();
REALMATRIX(const REALMATRIX& y);
// Copy Constructor: ---------------------------

REALMATRIX&   operator = (const REALMATRIX& y);

// Destructor: ---------------------------------

~REALMATRIX();
void clear();

// Access to matrix elements: ------------------

REAL&   operator ()           (unsigned int row,
                               unsigned int column) const;

REAL&   element               (unsigned int row,
                               unsigned int column) const;

// Standard Arithmetic: ------------------------

friend REALMATRIX  operator + (const REALMATRIX& x,
                               const REALMATRIX& y);
friend REALMATRIX  operator - (const REALMATRIX& x,
                               const REALMATRIX& y);
friend REALMATRIX  operator * (const REALMATRIX& x,
                               const REALMATRIX& y);
friend REALMATRIX  operator / (const REALMATRIX& x,
                               const REALMATRIX& y);

// Arithmetic with Scalar: ---------------------
friend REALMATRIX  operator * (const REALMATRIX& x,
                               const REAL& y);
inline friend REALMATRIX operator * (const REAL& y,
                               const REALMATRIX& x)
       {return x*y;};

friend REALMATRIX  operator / (const REALMATRIX& x,
                               const REAL& y);

// Information on Dimensions: ------------------

friend inline unsigned int rows (const REALMATRIX& x)
                {return x.maxrow;};
friend inline unsigned int columns (const REALMATRIX& x)
                {return x.maxcolumn;};

friend int bound (const REALMATRIX& x, 
                  const long k);

// Linear Algebra: -----------------------------

friend REALMATRIX eye     (unsigned int rows);
friend REALMATRIX zeroes  (unsigned int rows,
                           unsigned int columns);
friend REALMATRIX ones    (unsigned int rows,
                           unsigned int columns);
friend REAL       maxnorm (const REALMATRIX& x);
friend REAL       rowsumnorm (const REALMATRIX& x);
friend REAL       colsumnorm (const REALMATRIX& x);

friend REALMATRIX solve (
       REALMATRIX& lside,
       REALMATRIX& rside,
       int use_pivot);

// IO-routines: --------------------------

friend void       rwrite  (const REALMATRIX& x,
                           const long p,
                           const long w);
friend void       rshow   (const REALMATRIX& x,
                           const long w);

// Limits: --------------------------

friend REALMATRIX limit_lip (REALMATRIX f(long, const REALMATRIX&),
                           long lip,
                           const REALMATRIX& x);


public:
REAL*  values; 
unsigned int    maxrow,maxcolumn;
void adderror (sizetype error);
void seterror (sizetype error);
void geterror (sizetype& error) const;
};

#endif /* ! IRRAM_REALMATRIX_H */
