#ifndef IRRAM_SPARSEREALMATRIX_H
#define IRRAM_SPARSEREALMATRIX_H

#include "REALS.h" 

class SPM_ELEMENT
{ public:
REAL value;
unsigned int rowindex;
unsigned int colindex;
SPM_ELEMENT* nextcol;
SPM_ELEMENT* nextrow; 
};

class SPARSEREALMATRIX : public METRIC_OBJECT
{
public:
// Constructors: -------------------------------

SPARSEREALMATRIX(unsigned int rows,unsigned int columns);
SPARSEREALMATRIX();
SPARSEREALMATRIX(const SPARSEREALMATRIX& y);
// Copy Constructor: ---------------------------

SPARSEREALMATRIX&   operator = (const SPARSEREALMATRIX& y);

// Destructor: ---------------------------------

~SPARSEREALMATRIX();
void clear();
void free();

// Access to matrix elements: ------------------

friend void sparse_set(SPARSEREALMATRIX& m,unsigned int i,unsigned int j,const REAL& x);
friend void sparse_unset(SPARSEREALMATRIX& m,unsigned int i,unsigned int j);
friend SPM_ELEMENT* sparse_get_ptr(SPARSEREALMATRIX& m,unsigned int i,unsigned int j);

REAL    operator ()           (unsigned int row,
                               unsigned int column);


// Standard Arithmetic: ------------------------

friend SPARSEREALMATRIX  operator + (const SPARSEREALMATRIX& x,
                               const SPARSEREALMATRIX& y);
friend SPARSEREALMATRIX  operator - (const SPARSEREALMATRIX& x,
                               const SPARSEREALMATRIX& y);
friend SPARSEREALMATRIX  operator * (const SPARSEREALMATRIX& x,
                               const SPARSEREALMATRIX& y);
friend SPARSEREALMATRIX  operator / (const SPARSEREALMATRIX& x,
                               const SPARSEREALMATRIX& y);

// Arithmetic with Scalar: ---------------------
friend SPARSEREALMATRIX  operator * (const SPARSEREALMATRIX& x,
                               const REAL& y);
inline friend SPARSEREALMATRIX operator * (const REAL& y,
                               const SPARSEREALMATRIX& x)
       {return x*y;};

friend SPARSEREALMATRIX  operator / (const SPARSEREALMATRIX& x,
                               const REAL& y);

// Information on Dimensions: ------------------

friend inline unsigned int rows (const SPARSEREALMATRIX& x)
                {return x.maxrow;};
friend inline unsigned int columns (const SPARSEREALMATRIX& x)
                {return x.maxcolumn;};

friend int bound (const SPARSEREALMATRIX& x, 
                  const long k);

// Linear Algebra: --------------------------

friend SPARSEREALMATRIX sparse_eye     (unsigned int rows);
friend SPARSEREALMATRIX sparse_zeroes  (unsigned int rows,
                           unsigned int columns);
friend SPARSEREALMATRIX sparse_ones    (unsigned int rows,
                           unsigned int columns);
friend REAL       maxnorm (const SPARSEREALMATRIX& x);
friend REAL       rowsumnorm (const SPARSEREALMATRIX& x);
friend REAL       colsumnorm (const SPARSEREALMATRIX& x);

friend SPARSEREALMATRIX solve (
       SPARSEREALMATRIX& lside,
       SPARSEREALMATRIX& rside,
       int use_pivot);


// IO-routines: --------------------------

friend void       rwrite  (const SPARSEREALMATRIX& x,
                           const long p,
                           const long w);
friend void       rshow   (const SPARSEREALMATRIX& x,
                           const long w);

// Limits: --------------------------

friend SPARSEREALMATRIX limit_lip (SPARSEREALMATRIX f(long, const SPARSEREALMATRIX&),
                           long lip,
                           const SPARSEREALMATRIX& x);

//friend SPARSEREALMATRIX limit_mv (SPARSEREALMATRIX f(long prec,
//                                         long* choice,
//                                         const SPARSEREALMATRIX&),
//                           const SPARSEREALMATRIX& x);

//private:
SPM_ELEMENT**  rowvector; 
SPM_ELEMENT**  colvector; 
SPM_ELEMENT* hotspot;
unsigned int    maxrow,maxcolumn,filled;
void adderror (sizetype error);
void seterror (sizetype error);
void geterror (sizetype& error) const;
};


#endif /* ! IRRAM_SPARSEREALMATRIX_H */
