#include "GMP_ext.h"

#define   min(a,b)        (a < b ? a : b)
#define   max(a,b)        (a < b ? b : a)
#define   max3(a,b,c)       max(a,max(b,c))
#define   max4(a,b,c,d)       max(max(a,b),max(c,d))


long ext_gmp_var_count=0;

static ext_gmp_type gmp_FreeVars=0L;
static long gmp_FreeVarCount=0L;

void ext_gmp_initialize()
{
mpf_set_default_prec(32);
}

ext_gmp_type ext_gmp_init()
{ ext_gmp_type z;
  if (gmp_FreeVars) {
    z =gmp_FreeVars;
    gmp_FreeVars=gmp_FreeVars->_mp_exp;
    gmp_FreeVarCount -=1;
  } else {
  z = (ext_gmp_type) malloc(sizeof (__mpf_struct));
  mpf_init(z);
  }
  ext_gmp_var_count +=1;
  return z;
}

void ext_gmp_free(ext_gmp_type z)
{
  if ( gmp_FreeVarCount < 1000 ) {
    gmp_FreeVarCount +=1;
    z->_mp_exp=gmp_FreeVars;
    gmp_FreeVars=z;
  } else {
    mpf_clear(z); 
    free (z);
  }
  ext_gmp_var_count -=1;
}

long ext_gmp_size(ext_gmp_type z)
{
unsigned long r=0,w,n,m;
if (z->_mp_size==0) return GMP_min;
w=(z->_mp_d)[abs(z->_mp_size-1)];

n=w;
for (m=mp_bits_per_limb>>1; m>0; m=m>>1 ) {
      if (n>= (1<<m) ){
        r+=m; n=n>>m;
      }
    }
if (w > (1<<r) || abs(z->_mp_size) >0) r++;    

//for (i=1;i<=w;i+=i) {r+=1;if (r>=8*sizeof (mp_limb_t)) break;};

return r+(z->_mp_exp-1) *8* sizeof (mp_limb_t);
}

void ext_gmp_getsize(ext_gmp_type z,ext_gmp_sizetype* s)
{
  if ( z->_mp_size==0 ) {
    s->mantissa=0;
    s->exponent=GMP_min;
  } else if ( ((z->_mp_d)[abs(z->_mp_size)-1]==0) ) {
    s->mantissa=0;
    s->exponent=GMP_min;
  } else {
    s->mantissa=(z->_mp_d)[abs(z->_mp_size)-1];
    s->exponent=(z->_mp_exp-1) *8* sizeof (mp_limb_t);

    if ( s->mantissa <1<<(4* sizeof (mp_limb_t)) )
      { s->mantissa = s->mantissa <<(4* sizeof (mp_limb_t));
        s->exponent-=4* sizeof (mp_limb_t);
        if ( abs(z->_mp_size)>1 ) 
          s->mantissa+= (z->_mp_d)[abs(z->_mp_size)-2]
                         >>(4* sizeof (mp_limb_t));
    }
    if (s->mantissa > 1<<(8* sizeof (mp_limb_t)-1)) {
      s->mantissa=(s->mantissa>>1)+1;
      s->exponent+=1;
    }
    else 
    if (abs(z->_mp_size)>0) s->mantissa+=1;
  }
}

void ext_gmp_add(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,long p)
{ long q,s1,s2;
  s1=ext_gmp_size(z1);
  s2=ext_gmp_size(z2);
  q=max(s1,s2)-p+10;
  if (q<10)q=10;
  mpf_set_prec(z,q);
  mpf_add(z,z1,z2);
  return;
}

void ext_gmp_sub(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,long p)
{ long q,s1,s2;
  s1=ext_gmp_size(z1);
  s2=ext_gmp_size(z2);
  q=max(s1,s2)-p+10;
  if (q<10)q=10;
  mpf_set_prec(z,q);
  mpf_sub(z,z1,z2);
  return;
}

void ext_gmp_mul(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,long p)
{ long q,s1,s2;
  s1=ext_gmp_size(z1);
  s2=ext_gmp_size(z2);
  q=s1+s2-p+40;
  if (q<10)q=10;
  mpf_set_prec(z,q);
  mpf_mul(z,z1,z2);
  return;
}

void ext_gmp_div(ext_gmp_type z1,ext_gmp_type z2,ext_gmp_type z,long p)
{ long q,s1,s2;
  s1=ext_gmp_size(z1);
  s2=ext_gmp_size(z2);
  q=s1-s2-p+10;
  if (q<10)q=10;
  mpf_set_prec(z,q);
  mpf_div(z,z1,z2);
  return;
}

void ext_gmp_abs(ext_gmp_type z1,ext_gmp_type z)
{
  long q1=mpf_get_prec(z1);
  mpf_set_prec(z,q1);
  mpf_abs(z,z1);
  return;
}

void ext_gmp_truncate(ext_gmp_type z1,ext_gmp_type z)
{
  long q1=mpf_get_prec(z1);
  mpf_set_prec(z,q1);
  mpf_trunc(z,z1);
  return;
}


void ext_gmp_writee(ext_gmp_type z,long w)
{ char s[w];
  long e; 
  int i;
  int flag=0;
  mpf_get_str (s, &e, 10, w-7,z);
  for (i=0;i<w-7;i++) {
    if ( s[i]==0 ) flag=1 ;
    if ( flag ) s[i]='0';
  } 
  if (s[0]=='-') {s[w-7]=0;printf("-.%sE%+05ld",&s[1],e);}
  else {s[w-8]=0;printf("+.%sE%+05ld",&s[0],e);};
}

void ext_gmp_write(ext_gmp_type z,long w)
{ char s[w];
  long e; 
  int i;
  int flag=0;
  mpf_get_str (s, &e, 10, w-7,z);
  for (i=0;i<w-7;i++) { if ( s[i]==0 ) flag=1 ;
    if ( flag ) s[i]='0';
  } 
  if (s[0]=='-') {s[w-7]=0;printf("-.%sE%+05ld",&s[1],e);}
  else {s[w-8]=0;printf("+.%sE%+05ld",&s[0],e);};
  if (z->_mp_size != 0) {
     printf("\n");
     for (i=abs(z->_mp_size-1);i>=0; i-=1) 
       printf("%lu ",z->_mp_d[i]);
  }
  printf("\nExponent %ld, Limbs: %d, Bits: %ld\n",
     z->_mp_exp,z->_mp_size,mpf_get_prec(z));
}

void ext_gmp_duplicate_wo_init(ext_gmp_type z1,ext_gmp_type z2)
{
  long q1=mpf_get_prec(z1);
  if (mpf_get_prec(z2)< q1) mpf_set_prec(z2,q1);
  mpf_set(z2,z1);
}

void ext_gmp_duplicate_w_init(ext_gmp_type z1,ext_gmp_type *z2)
{
//  *z2 = (ext_gmp_type) malloc(sizeof (__mpf_struct));
//  mpf_init2(*z2,mpf_get_prec(z1));
  long q1=mpf_get_prec(z1);
  *z2=ext_gmp_init();
  if (mpf_get_prec(*z2)< q1) mpf_set_prec(*z2,q1);
  ext_gmp_var_count +=1;
  mpf_set(*z2,z1);
}

void ext_gmp_sqrt(ext_gmp_type z1,ext_gmp_type z,long p)
{ long q,s1,s2;
  s1=ext_gmp_size(z1);
  q=max(s1,s2)-p+10;
  if (q<10)q=10;
  mpf_set_prec(z,q);
  mpf_sqrt(z,z1);
  return;
}

void ext_gmp_shift(ext_gmp_type z1,ext_gmp_type z,long n)
{
  mpf_set_prec(z,mpf_get_prec(z1));
  if ( n>=  0 ) mpf_mul_2exp(z,z1,n);
  else mpf_div_2exp(z,z1,-n);
  return;
}

