#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "SYSTEM_.h"

#ifndef DEFINITION_Dyadic_Base
#include "Dyadic_Base.h"
#endif

#ifndef DEFINITION_Dyadic
#include "Dyadic.h"
#endif

#define   max(a,b)        (a < b ? b : a)
#define   min(a,b)        (a < b ? a : b)

Dyadic_LRErrorProc Dyadic_LRError;
unsigned long Dyadic_LRVarCount;

#define BLKFACTOR	10240
/* BLKFACTOR: Number of blocks allocated at once if list of free blocks
   is empty, slightly reduces memory consumption.
*/

typedef Dyadic_Base_WertFeld WertFeld;

static WertFeld AKKUMULATOR1;
static WertFeld AKKUMULATOR2;
static WertFeld AKKUMULATOR3;
static unsigned int MSB [Dyadic_Base_BITZAHL];
static double basis;
static double lnbasis;
static unsigned long LRSpaceLow;
static unsigned long LRmaxSpaceLow;
static Dyad *CONV_HILFE, *LRINT_ARG, *A_LRSQRT, *B_LRSQRT;
static long trigger;
static void Initialize ();
void LRSET (Dyad* *Z, long HighExp, Dyadic_Base_WertFeld AKKU, long Anzahl, BOOLEAN NEGATIV);
void LRGET (const Dyad *Z, long HighExp, Dyadic_Base_WertFeld AKKU, long Anzahl);
static void LRTODEC (const Dyad * Z, long A, unsigned char Dez[], unsigned long , long *DExp);
static void SPLIT (const Dyad* X, Dyad* *Y, Dyad*  *Z, long s);
static void LICOPY (const Dyad* X, Dyad* *Z);
inline BlockZeiger LRNewBlock();
inline void LRFreeBlock(BlockZeiger q);
 
static BlockZeiger LRFreeBlocks=0L;
static BlockZeiger LRDummyBlock;
static Dyad* LRFreeVars=0L;

inline BlockZeiger LRNewBlock()
{
    BlockZeiger p;
    LRSpaceLow++;
    LRmaxSpaceLow=max(LRmaxSpaceLow,LRSpaceLow);
    if (! LRFreeBlocks)  {
      int i;
      BlockZeiger a;
      a = malloc( BLKFACTOR *sizeof(Block));
      LRFreeBlocks = a;
      for (i=1; i< BLKFACTOR; i++) {
        a->BlockNext = a+1;
        a++;
        }
      a->BlockNext=0L;
      }
    p= LRFreeBlocks;
    LRFreeBlocks=p->BlockNext;
    return p;
}

inline void LRFreeBlock(BlockZeiger q)
{
    q->BlockNext =LRFreeBlocks;
    LRFreeBlocks=q;
    LRSpaceLow--;
}

static void Initialize()
{
  long i;
  unsigned int d;

  d = 1;
  basis = 1.0;
  for (i = 1; i <= Dyadic_Base_BITZAHL; i += 1) {
    MSB[i - 1] = d;
    basis = 2.0 * basis;
    d += d;
  }
  lnbasis = log(basis);
  LRDummyBlock=LRNewBlock();
  Dyadic_LRError = Dyadic_LRErrorStatus;
  Dyadic_LRINIT(&CONV_HILFE);
  Dyadic_LRINIT(&LRINT_ARG);
  Dyadic_LRINIT(&A_LRSQRT);
  Dyadic_LRINIT(&B_LRSQRT);
  Dyadic_LRVarCount = 0;
  LRSpaceLow = 0;
  LRmaxSpaceLow = 0;
  trigger = 100;
/*  printf("trigger fuer Karatsuba :\n");
  scanf("%d",&trigger);
  printf("\n");*/
}

void Dyadic_LRErrorIgnore (int Error)
{
}

void Dyadic_LRErrorStatus (int Error)
{
  fprintf(stderr,"Dyadic-Error, ");
  switch (Error) {
  case 1:;
    fprintf(stderr,"LRTOINT-Overflow\n");
    break;
  case 2:;
    fprintf(stderr,"Division durch 0\n");
    break;
  case 3:;
    fprintf(stderr,"Falscher CHAR bei DECTOLR\n");
    break;
  case 4:;
    fprintf(stderr,"Negatives Argument bei LRSQRT\n");
    break;
  default :
    fprintf(stderr,"Fehler %d gefunden\n",Error);
    break;
  }
}

void Dyadic_LRErrorHalt (int Error)
{
  Dyadic_LRErrorStatus(Error);
  exit(1);
}

unsigned long Dyadic_LRmaxSpace ()
{
  return LRmaxSpaceLow / 1024 * sizeof(Block);
}

unsigned long Dyadic_LRSpace ()
{
  return LRSpaceLow / 1024 * sizeof(Block);
}

unsigned long Dyadic_LOG2 (unsigned long N)
{
  unsigned long L;

  L = 0;
  for (;;) {
    if (N < 2) {
      goto EXIT_1;
    }
    L++;
    N = N / 2;
  } EXIT_1:;
  return L;
}

void Dyadic_LRCLEAR (Dyad* *Z)
{
  if ( (*Z)->reference_count ==1 ) {
    BlockZeiger p, q;

    p = (*Z)->Wert; 
    while (p != 0L) {
      q = p;
      p = p->BlockNext;
      LRFreeBlock(q);
    }
    (*Z)->Wert=LRFreeVars;
    LRFreeVars=(*Z);
    (*Z)=0L;
    Dyadic_LRVarCount -= 1;
  } else { 
 (*Z)->reference_count -= 1;
 }
}

void Dyadic_LRINIT (Dyad* *Z)
{
  if (LRFreeVars) {
    (*Z) =LRFreeVars;
    LRFreeVars=LRFreeVars->Wert;
  } else {
  (*Z)=malloc(sizeof(Dyad));
  }
  Dyadic_LRVarCount++;
  (*Z)->Wert = 0L;
  (*Z)->reference_count=1;
}

void Dyadic_LRNEGATE (Dyad* *Z)
{
  if ( (*Z)->reference_count > 1 ) Dyadic_LRDUPLICATE(*Z,Z);
  (*Z)->negativ = !(*Z)->negativ;
}

void Dyadic_LRABSVAL (Dyad* *Z)
{
  if ( (*Z)->reference_count > 1  ) Dyadic_LRDUPLICATE(*Z,Z);
  (*Z)->negativ = FALSE;
}

void LRSET
(Dyad* *Z, long HighExp, Dyadic_Base_WertFeld AKKU, long Anzahl, BOOLEAN NEGATIV)
{
  long m, MinAkku, MaxAkku;
  BlockZeiger p, q;
  unsigned int *MaxP,*MinP,*WertP;

  if ((*Z)->reference_count > 1) {
    (*Z)->reference_count -=1;
    Dyadic_LRINIT(Z);
  }
  MaxP = &AKKU[Anzahl - 1];
  MaxAkku = Anzahl;
  while (*MaxP == 0 && MaxP > AKKU) {
    MaxP--;
    MaxAkku--;
  }
  MinP = AKKU;
  MinAkku = 1;
  while (*MinP  == 0 && MaxP > MinP) {
    MinP++;
    MinAkku++;
  }
  if ( (*Z)->Wert == 0L ) {
    p=LRNewBlock();
    (*Z)->Wert = p;
    p->BlockNext = 0L;
  } else {
    p = (*Z)->Wert;
  }
  m = BLOCKSIZE + 1;
  WertP= &(p->BlockFeld[BLOCKSIZE-1]);

  while (MinP <= MaxP ) {
      m--;
      if (m < 1) {
        if (p->BlockNext == 0L) {
          p->BlockNext=LRNewBlock();
          p = p->BlockNext;
          p->BlockNext = 0L;
        } else {
          p = p->BlockNext;
        }
        m = BLOCKSIZE;
        WertP= &(p->BlockFeld[BLOCKSIZE-1]);
      }
      *(WertP--) = *(MinP++);
    }

  q = p->BlockNext;
  p->BlockNext = 0L;
  while (q != 0L) {
    p = q;
    q = q->BlockNext;
    LRFreeBlock(p);
  }
  {
    (*Z)->anzahl = MaxAkku - MinAkku + 1;
    if ((*Z)->Wert->BlockFeld[BLOCKSIZE-1] == 0)
      (*Z)->highexponent=Dyadic_Base_MinInt;
    else
      (*Z)->highexponent = HighExp + 1 - MinAkku;
    (*Z)->negativ = NEGATIV;
  }
}

void LRGET
(const Dyad * Z, long HighExp, Dyadic_Base_WertFeld AKKU, long Anzahl)
{
  long m;
  long IndAkku, MinAkku, MaxAkku;
  unsigned int *AP;
  BlockZeiger p;

  if (Anzahl >=  Dyadic_Base_MAXSIZE) {
     fprintf(stderr,"Akkumulator-Groesse ueberschritten!\n");
     exit(1);
   } 
  {

    if (Z->highexponent < HighExp - Anzahl) {
      MinAkku = Anzahl;
      MaxAkku = Anzahl;
    } else {
      MinAkku = HighExp - Z->highexponent;
      if (Anzahl > MinAkku + Z->anzahl) {
        MaxAkku = MinAkku + Z->anzahl;
      } else {
        MaxAkku = Anzahl;
      }
    }

    AP = AKKU;
    IndAkku=1;
    while ( IndAkku++ <= MinAkku ) *(AP++) = 0;

    m = BLOCKSIZE + 1;
    p = Z->Wert;

    IndAkku=MinAkku+1;
    while ( IndAkku++ <= MaxAkku )
    {
      m--;
      if (m < 1) {
        m = BLOCKSIZE;
        p = p->BlockNext;
      }
      *(AP++) = p->BlockFeld[m - 1];
    }

    IndAkku=MaxAkku+1;
    while ( IndAkku++ <= Anzahl ) *(AP++) = 0;

  }
}

void Dyadic_INTTOLR
(long W, Dyad* *Z)
{
  long I;
  unsigned long J, N, M, B;

  N = 1;
  I = W;
  if (I < 0) {
    M =  -I;
  } else {
    M =  I;
  }
  B = M;
  while (B >= BASIS) {
    N++;
    B = B / BASIS;
  }
  for (J = N; J >= 1; J += -1) {
    AKKUMULATOR1[J - 1] = M % BASIS;
    M = M / BASIS;
  }
  LRSET(Z, N, AKKUMULATOR1, N, W < 0);
}

void Dyadic_LRWRITE
(const Dyad * Z, long F)
{
  unsigned char  BIT [Dyadic_Base_BITZAHL];
  long i, j, n;
  unsigned long l;
  BlockZeiger p;
  long length;

  printf("HighExponent: %5ld",Z->highexponent);
  printf(" Anzahl: %5ld",Z->anzahl);
  printf(" Vorzeichen: ");
  if ( Z->negativ ) {
    printf("negativ");
  } else {
    printf("positiv");
  }
  printf("\n");
  p = Z->Wert;
  n = BLOCKSIZE + 1;
  length = 64;
  for (i = Z->anzahl; i >= 1; i += -1) {
    n--;
    if (n < 1) {
      p = p->BlockNext;
      n = BLOCKSIZE;
    }
    l = p->BlockFeld[n - 1];
    for (j = 1; j <= Dyadic_Base_BITZAHL; j += 1) {
      if (l % 2 != 0) {
        BIT[j - 1] = '1';
      } else {
        BIT[j - 1] = '0';
      }
      l = l / 2;
    }
    for (j = Dyadic_Base_BITZAHL; j >= 1; j += -1) {
      printf("%c",BIT[j - 1]);
      F++;
      if (F == 0) {
        printf("\n");
        return;
      }
      length--;
      if (length == 0) {
        printf("\n");
        length = 64;
      }
    }
  }
  for (i = F; i <= 0; i += 1) {
    printf("0");
    length--;
    if (length == 0) {
      printf("\n");
      length = 64;
    }
  }
  if (length != 64) {
    printf("\n");
  }
}

void Dyadic_LRTOREAL
(const Dyad * X, double *R)
{
  double e;

  LRGET(X, X->highexponent, AKKUMULATOR1, 3);
  if ( X->highexponent < 0) {
    e = -FLOAT(VAL(unsigned long, -X->highexponent)) * lnbasis;
  } else {
    e = FLOAT(VAL(unsigned long, X->highexponent)) * lnbasis;
  }
  if (e < -20.0 || AKKUMULATOR1[0] == 0) {
    *R = 0.0;
  } else {
    *R = ((FLOAT(AKKUMULATOR1[2]) / basis + FLOAT(AKKUMULATOR1[1])) / basis + FLOAT(AKKUMULATOR1[0])) / basis * exp(e);
    if ( X->negativ) {
      *R = -*R;
    }
  }
}

void Dyadic_LRTOINT
(const Dyad * X, long *I)
{
  unsigned long e;

  e = X->highexponent;
  LRGET(X, e, AKKUMULATOR1, 1);
  if (AKKUMULATOR1[0] == 0) {
    *I = 0;
  } else if (e > 1) {
    *I = 0;
    (*Dyadic_LRError)(1);
  } else if (e < 1) {
    *I = 0;
  } else {
    *I = AKKUMULATOR1[0];
  }
  if ( X->negativ) {
    *I = -*I;
  }
}

static void LRTODEC
(const Dyad *Z, long Size, unsigned char Dez[], unsigned long O_1, long *DExp)
{
  long R, L, i, j, n, m;
  unsigned long k;
  unsigned int NULLWERT;
  long Groesse;
  unsigned long * Dezi;

  Dezi=malloc(O_1*sizeof(unsigned long));

  Groesse = O_1-2;
  if (Groesse < Size) {
    L = Groesse;
  } else {
    L = Size;
  }

  NULLWERT = ORD('0');
  if ( Z->highexponent > 0) {
    n = Z->highexponent;
    m = n;
    LRGET(Z, m, AKKUMULATOR1, m);
    *DExp = 0;
    while (m > 0) {
      AKKUMULATOR3[0] = 10;
      Dyadic_Base_SysDiv(AKKUMULATOR1, AKKUMULATOR3, AKKUMULATOR2, m, 1, m, TRUE);
      Dezi[*DExp % Groesse] = AKKUMULATOR1[0];
      (*DExp)++;
      if (AKKUMULATOR2[0] == 0) {
        m--;
        {
          long B_9 = 1, B_10 = m;

          if (B_9 <= B_10)
            for (j = B_9;; j += 1) {
              AKKUMULATOR1[j - 1] = AKKUMULATOR2[j];
              if (j >= B_10) break;
            }
        }
      } else {
        {
          long B_11 = 1, B_12 = m;

          if (B_11 <= B_12)
            for (j = B_11;; j += 1) {
              AKKUMULATOR1[j - 1] = AKKUMULATOR2[j - 1];
              if (j >= B_12) break;
            }
        }
      }
    }
    if (*DExp <= L) {
      R = L - *DExp;
      m = 10 * R / (3 * Dyadic_Base_BITZAHL) + 3;
      LRGET(Z, n, AKKUMULATOR2, n + m);
      {
        long B_13 = 1, B_14 = m;

        if (B_13 <= B_14)
          for (j = B_13;; j += 1) {
            AKKUMULATOR1[j - 1] = AKKUMULATOR2[n + j - 1];
            if (j >= B_14) break;
          }
      }
      AKKUMULATOR3[0] = 10;
      {
        long B_15 = 1, B_16 = R + 1;

        if (B_15 <= B_16)
          for (i = B_15;; i += 1) {
            Dyadic_Base_SysMul(AKKUMULATOR1, AKKUMULATOR3, AKKUMULATOR2, m, 1, m + 2);
            Dezi[Groesse - i] = AKKUMULATOR2[0];
            {
              long B_17 = 2, B_18 = m + 1;

              if (B_17 <= B_18)
                for (j = B_17;; j += 1) {
                  AKKUMULATOR1[j - 1 - 1] = AKKUMULATOR2[j - 1];
                  if (j >= B_18) break;
                }
            }
            if (i >= B_16) break;
          }
      }
    }
  } else {
    n = Z->highexponent;
    *DExp = 0;
    m = 10 * L / (3 * Dyadic_Base_BITZAHL) + 3;
    LRGET(Z, n, AKKUMULATOR1, m);
    AKKUMULATOR3[0] = 10;
    for (;;) {
      Dyadic_Base_SysMul(AKKUMULATOR1, AKKUMULATOR3, AKKUMULATOR2, m, 1, m + 2);
      if (AKKUMULATOR2[m] == 0) {
        m--;
      }
      if (m == 0) {
        goto EXIT_2;
      }
      if (AKKUMULATOR2[0] != 0) {
        if (n == 0) {
          goto EXIT_2;
        }
        {
          long B_19 = 1, B_20 = m + 1;

          if (B_19 <= B_20)
            for (j = B_19;; j += 1) {
              AKKUMULATOR1[j - 1] = AKKUMULATOR2[j - 1];
              if (j >= B_20) break;
            }
        }
        n++;
        m++;
      } else {
        {
          long B_21 = 2, B_22 = m + 1;

          if (B_21 <= B_22)
            for (j = B_21;; j += 1) {
              AKKUMULATOR1[j - 1 - 1] = AKKUMULATOR2[j - 1];
              if (j >= B_22) break;
            }
        }
      }
     (*DExp)--;
    } EXIT_2:;
    k = (Groesse + (*DExp - 1) % Groesse) % Groesse;
    Dezi[k] = AKKUMULATOR2[0];
    {
      long B_23 = 2, B_24 = m + 1;

      if (B_23 <= B_24)
        for (j = B_23;; j += 1) {
          AKKUMULATOR1[j - 1 - 1] = AKKUMULATOR2[j - 1];
          if (j >= B_24) break;
        }
    }
    {
      if (1 <= L)
        for (i = 1;; i += 1) {
          Dyadic_Base_SysMul(AKKUMULATOR1, AKKUMULATOR3, AKKUMULATOR2, m, 1, m + 2);
          k = (Groesse + k - 1) % Groesse;
          Dezi[k] = AKKUMULATOR2[0];
          {
             if (2 <= m+1)
              for (j = 2;; j += 1) {
                AKKUMULATOR1[j - 2] = AKKUMULATOR2[j - 1];
                if (j >= m+1) break;
              }
          }
          if (i >= L) break;
        }
    }
  }
  if ( Z->negativ) {
    Dez[0] = '-';
  } else {
    Dez[0] = '+';
  }
  Dez[1] = '.';  i = 1;
  {
      for (j = *DExp - 1;; j += -1) {
        i++;
        Dez[i] = CHR(NULLWERT + Dezi[(Groesse + j % Groesse) % Groesse]);
        if (j <= *DExp - L) break;
      }
  }
  if (Dez[L] >= CHR(NULLWERT + 6) ) {
     for (i=L-1;i>1;i--) {
       if (Dez[i] != '9' ) {
         Dez[i]=Dez[i]+1;
         break;
       } else Dez[i]='0';
     }
     if (i==1) {Dez[2]='1';(*DExp)++;};
  }
//  Dez[1] = Dez[2];
//  Dez[2]='.';
//  (*DExp)--;
  Dez[L] = CHR(0);
  free(Dezi);
}

void Dyadic_DECTOLR
(const unsigned char Dez[], unsigned long O_2, long DExp, Dyad* * Z)
{
  unsigned long i;
  long j;
  OPEN_ARRAY_LOCALS

  ALLOC_OPEN_ARRAYS(O_2 * sizeof(unsigned char), 1)
  COPY_OPEN_ARRAY(Dez, O_2, unsigned char)
  Dyadic_INTTOLR(0, Z);
  i = 0;
  for (;;) {
    if (i == (O_2 - 1)) {
      goto EXIT_3;
    }
    if (Dez[i] == CHR(0)) {
      goto EXIT_3;
    }
    if (Dez[i] < '0' || Dez[i] > '9') {
      (*Dyadic_LRError)(3);
      j = 0;
    } else {
      j = ORD(Dez[i]) - ORD('0');
    }
    Dyadic_INTTOLR(j, &CONV_HILFE);
    Dyadic_LRMULTINT(*Z, 10, Z, -2);
    Dyadic_LRADD(*Z, CONV_HILFE, Z, -2);
    i++;
  } EXIT_3:;
  if (DExp < 0) {
    Dyadic_INTTOLR(10, &CONV_HILFE);
    {
      long B_31 = 2, B_32 = -DExp;

      if (B_31 <= B_32)
        for (j = B_31;; j += 1) {
          Dyadic_LRMULTINT(CONV_HILFE, 10, &CONV_HILFE, -2);
          if (j >= B_32) break;
        }
    }
    Dyadic_LRDIV(*Z, CONV_HILFE, Z, -Dyadic_LRSIZE(CONV_HILFE));
  } else {
    {
      long B_33 = 1, B_34 = DExp;

      if (B_33 <= B_34)
        for (j = B_33;; j += 1) {
          Dyadic_LRMULTINT(*Z, 10, Z, -2);
          if (j >= B_34) break;
        }
    }
  }
  FREE_OPEN_ARRAYS
}

void Dyadic_LRWRITEE
(const Dyad * Z, long L)
{
  long DExp;
  unsigned char * Dez;
  Dez=malloc(L*sizeof(unsigned char));

  if (L < 9) {
    printf("*********");
  } else {
    LRTODEC(Z, L - 6, Dez, L, &DExp);
    Dez[L-6]=0;
    printf("%s",Dez);
    printf("E");
    if (DExp < 0) {
      printf("-");
      DExp = -DExp;
    } else {
      printf("+");
    }
    printf("%04ld",DExp);
  }
  free(Dez);
}

void Dyadic_LRWRITEF
(const Dyad * Z, long L, long R)
{
  long DExp, n, i, j, m;
  unsigned char * Dez;

  Dyadic_INTTOLR(5, &CONV_HILFE);
  

  if (0 <= R)
    for (i = 0; i<= R ; i += 1) {
      Dyadic_LRDIVINT(CONV_HILFE, 10, &CONV_HILFE, -R * 4);
     }

  if (Dyadic_LRSIGN(Z) == -1) {
    Dyadic_LRSUB(Z, CONV_HILFE, &CONV_HILFE, -R * 4);
  } else {
    Dyadic_LRADD(Z, CONV_HILFE, &CONV_HILFE, -R * 4);
  }
  Dez=malloc(L*sizeof(unsigned char));
  LRTODEC(CONV_HILFE, L, Dez, L, &DExp);
  if (L - R - 2 >= DExp) {
    if (DExp <= 0) {
      n = 0;
      m = -DExp;
    } else {
      n = DExp;
      m = 0;
    }

    if (n + 1 <= L - R - 2)
      for (i = n + 1 ;; i += 1) {
          printf(" ");
          if (i >= L - R - 2) break;
        }

    printf("%c",Dez[0]);

    for (i = 1 ; i <= n ; i += 1) {
          printf("%c",Dez[i + 1]);
        }
    printf(".");
    if (m > R) {
      m = R;
    }
    for (i = 1 ; i <= m ; i += 1)  printf("0");

    {
      long B_47 = n + 1, B_48 = n + R - m;

      if (B_47 <= B_48)
        for (i = B_47;; i += 1) {
          printf("%c",Dez[i + 1]);
          if (i >= B_48) break;
        }
    }
  } else {
    if (L < 9) {
      {
        long B_49 = 1, B_50 = L;

        if (B_49 <= B_50)
          for (j = B_49;; j += 1) {
            printf("*");
            if (j >= B_50) break;
          }
      }
    } else {
      {
        long B_51 = 1, B_52 = L - 6;

        if (B_51 <= B_52)
          for (j = B_51;; j += 1) {
            printf("%c",Dez[j - 1]);
            if (j >= B_52) break;
          }
      }
      printf("E");
      if (DExp < 0) {
        printf("-");
        DExp = -DExp;
      } else {
        printf("+");
      }
      j = 1000;
      for (i = 1; i <= 4; i += 1) {
        printf("%1ld",DExp / j);
        DExp = DExp % j;
        j = j / 10;
      }
    }
  }
  free(Dez);
}

void Dyadic_LRADD
(const Dyad * X, const Dyad * Y, Dyad* *Z, long N)
{
  long Anzahl, HighExp;
  long n;
  BOOLEAN AkkuNegativ;

  if (N > 0) {
    n = (N - 1) / Dyadic_Base_BITZAHL;
  } else {
    n = N / Dyadic_Base_BITZAHL;
    n--;
  }
  if ( X->highexponent > Y->highexponent) {
    HighExp = X->highexponent;
  } else {
    HighExp = Y->highexponent;
  }
  if (n > HighExp) {
    Dyadic_INTTOLR(0, Z);
    return;
  }
  if ( X->highexponent - X->anzahl > Y->highexponent - Y->anzahl) {
    Anzahl = HighExp - Y->highexponent + Y->anzahl;
  } else {
    Anzahl = HighExp - X->highexponent + X->anzahl;
  }
  if (Anzahl > HighExp - n) {
    Anzahl = HighExp - n;
  }
  LRGET(X, HighExp, AKKUMULATOR1, Anzahl);
  LRGET(Y, HighExp, AKKUMULATOR2, Anzahl);
  if ( X->negativ == Y->negativ) {
    Dyadic_Base_SysAdd(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, Anzahl);
    LRSET(Z, HighExp + 1, AKKUMULATOR3, Anzahl + 1, X->negativ);
  } else {
    Dyadic_Base_SysSub(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, Anzahl, &AkkuNegativ);
    if (AkkuNegativ) {
      LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, !X->negativ);
    } else {
      LRSET(Z, HighExp, AKKUMULATOR3, Anzahl,  X->negativ);
    }
  }
}

void Dyadic_LRSUB
(const Dyad * X, const Dyad * Y, Dyad* * Z, long N)
{
  long Anzahl, HighExp;
  long n;
  BOOLEAN AkkuNegativ;

  if (N > 0) {
    n = (N - 1) / Dyadic_Base_BITZAHL;
  } else {
    n = N / Dyadic_Base_BITZAHL;
    n--;
  }
  if ( X->highexponent > Y->highexponent) {
    HighExp = X->highexponent;
  } else {
    HighExp = Y->highexponent;
  }
  if (n > HighExp) {
    Dyadic_INTTOLR(0, Z);
    return;
  }
  if ( X->highexponent - X->anzahl > Y->highexponent - Y->anzahl) {
    Anzahl = HighExp - Y->highexponent + Y->anzahl;
  } else {
    Anzahl = HighExp - X->highexponent + X->anzahl;
  }
  if (Anzahl > HighExp - n) {
    Anzahl = HighExp - n;
  }
  LRGET(X, HighExp, AKKUMULATOR1, Anzahl);
  LRGET(Y, HighExp, AKKUMULATOR2, Anzahl);
  if ( X->negativ != Y->negativ) {
    Dyadic_Base_SysAdd(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, Anzahl);
    LRSET(Z, HighExp + 1, AKKUMULATOR3, Anzahl + 1, X->negativ);
  } else {
    Dyadic_Base_SysSub(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, Anzahl, &AkkuNegativ);
    if (AkkuNegativ) {
      LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, !X->negativ);
    } else {
      LRSET(Z, HighExp, AKKUMULATOR3, Anzahl,  X->negativ);
    }
  }
}

void Dyadic_LRMULT
(const Dyad * X, const Dyad * Y, Dyad* * Z, long N)
{
  long HighExp, Anzahl, XAnz, YAnz;
  long n;

  if (N - 2 >= 0) {
    n = (N - 2) / Dyadic_Base_BITZAHL;
  } else {
    n = (N - 1) / Dyadic_Base_BITZAHL;
    n--;
  }
  HighExp = X->highexponent + Y->highexponent;
  if (n >= HighExp) {
    Dyadic_INTTOLR(0, Z);
    return;
  }
  Anzahl = HighExp - n;
  XAnz = X->anzahl;
  if (XAnz > Anzahl) {
    XAnz = Anzahl;
  }
  LRGET(X, X->highexponent, AKKUMULATOR1, XAnz);
  YAnz = Y->anzahl;
  if (YAnz > Anzahl) {
    YAnz = Anzahl;
  }
  LRGET(Y, Y->highexponent, AKKUMULATOR2, YAnz);
  Dyadic_Base_SysMul(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, XAnz, YAnz, Anzahl);
  LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, X->negativ != Y->negativ);
}

void Dyadic_LRMULTKARAT (const Dyad * X, const Dyad * Y, Dyad* *Z, long N)
{
  Dyad *XC;
  long HighExp, Anzahl, XAnz, YAnz;
  long n;

  if ((Dyadic_LRSIGN(X) == 0) || Dyadic_LRSIGN(Y) == 0){ 
    Dyadic_INTTOLR(0, Z);
    return;
  }
  if (N - 2 >= 0) {
    n = (N - 2) / Dyadic_Base_BITZAHL;
  } else {
    n = (N - 1) / Dyadic_Base_BITZAHL;
    DEC(n);
  }
  HighExp = X->highexponent + Y->highexponent;
  if (n >= HighExp) {
    Dyadic_INTTOLR(0, Z);
    return;
  }
  Anzahl = HighExp - n;
  XAnz = X->anzahl;
  if (XAnz > Anzahl) {
    XAnz = Anzahl;
  }
  YAnz = Y->anzahl;
  if (YAnz > Anzahl) {
    YAnz = Anzahl;
  }
  if (Anzahl < 4 * trigger) {
    LRGET(X, X->highexponent, AKKUMULATOR1, XAnz);
    LRGET(Y, Y->highexponent, AKKUMULATOR2, YAnz);
    Dyadic_Base_SysMul(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, XAnz, YAnz, Anzahl);
    LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, X->negativ != Y->negativ);
  } else {
    Dyadic_LRINIT(&XC);
    LRGET(X, X->highexponent, AKKUMULATOR1, XAnz);
    LRSET(&XC, XAnz, AKKUMULATOR1, XAnz, X->negativ);
    LRGET(Y, Y->highexponent, AKKUMULATOR1, YAnz);
    LRSET(Z, YAnz, AKKUMULATOR1, YAnz, Y->negativ);
    Dyadic_LIMULTKARAT(XC, *Z, Z);
    (*Z)->highexponent = (*Z)->highexponent + HighExp - XAnz - YAnz;
    Dyadic_LRCLEAR(&XC);
  }
}

void Dyadic_LRDIV
(const Dyad * X, const Dyad * Y, Dyad* * Z, long N)
{
  long YAnzahl, Anzahl;
  long n;

  if (Dyadic_LRSIGN(Y) == 0) {
    Dyadic_INTTOLR(0, Z);
    (*Dyadic_LRError)(2);
    return;
  }
  if (N - 2 > 0) {
    n = (N - 2) / Dyadic_Base_BITZAHL;
  } else {
    n = (N - 1) / Dyadic_Base_BITZAHL;
    n--;
  }
  Anzahl = X->highexponent - Y->highexponent - n + 1;
  if (Anzahl <= 0) {
    Dyadic_INTTOLR(0, Z);
  } else {
    LRGET(X, X->highexponent, AKKUMULATOR1, Anzahl);
    if (Y->anzahl < Anzahl + 1) {
      YAnzahl = Y->anzahl;
    } else {
      YAnzahl = Anzahl + 1;
    }
    LRGET(Y, Y->highexponent, AKKUMULATOR2, YAnzahl);
    Dyadic_Base_SysDiv(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, Anzahl, YAnzahl, Anzahl, FALSE);
    LRSET(Z, 1 + X->highexponent - Y->highexponent, AKKUMULATOR3, Anzahl, X->negativ != Y->negativ);
  }
}

void Dyadic_LRSHIFT
(const Dyad * X, Dyad* *Z, long I)
{
  long n;
  unsigned long ShiftZahl;
  unsigned long Anzahl;

  Anzahl = X->anzahl + 1;
  LRGET(X, X->highexponent + 1, AKKUMULATOR1, Anzahl);
  if (I >= 0) {
    n = I / Dyadic_Base_BITZAHL;
  } else {
    n = (I + 1) / Dyadic_Base_BITZAHL;
    n--;
  }
  ShiftZahl = I - n * Dyadic_Base_BITZAHL;
  if (ShiftZahl != 0) {
    Dyadic_Base_SysBlockShl(AKKUMULATOR1, Anzahl, ShiftZahl);
  }
  LRSET(Z, X->highexponent + 1 + n, AKKUMULATOR1, Anzahl, X->negativ);
}

void Dyadic_LRSQUARE
(const Dyad * X, Dyad* *Z, long N)
{
  long HighExp, Anzahl, XAnz;
  long n;

  if (N >= 2) {
    n = (N - 2) / Dyadic_Base_BITZAHL;
  } else {
    n = (N - 1) / Dyadic_Base_BITZAHL;
    n--;
  }
  HighExp = X->highexponent + X->highexponent;
  if (n >= HighExp) {
    Dyadic_INTTOLR(0, Z);
    return;
  }
  Anzahl = HighExp - n;
  XAnz = X->anzahl;
  if (XAnz > Anzahl) {
    XAnz = Anzahl;
  }
  LRGET(X, X->highexponent, AKKUMULATOR1, XAnz);
  Dyadic_Base_SysMul(AKKUMULATOR1, AKKUMULATOR1, AKKUMULATOR3, XAnz, XAnz, Anzahl);
  LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, FALSE);
}

void Dyadic_LRMULTINT
(const Dyad * X, long Y, Dyad* * Z, long N)
{
  Dyadic_INTTOLR(Y, &LRINT_ARG);
  Dyadic_LRMULT(X, LRINT_ARG, Z, N);
}

void Dyadic_LRDIVINT
(const Dyad * X, long Y, Dyad* *Z, long N)
{
  Dyadic_INTTOLR(Y, &LRINT_ARG);
  Dyadic_LRDIV(X, LRINT_ARG, Z, N);
}

long Dyadic_LRSIGN
(const Dyad * X)
{
  if (X->Wert->BlockFeld[BLOCKSIZE - 1] == 0) {
    return 0;
  } else {
    if (X->negativ) {
      return -1;
    } else {
      return 1;
    }
  }
}

void Dyadic_LRGETSIZE
(const Dyad * X,LRsizetype* s)
{
  unsigned long d;
  s->mantissa = X->Wert->BlockFeld[BLOCKSIZE - 1];
  if (s->mantissa == 0) {
    s->exponent=Dyadic_Base_MinInt;
  } else {
    s->exponent=Dyadic_Base_BITZAHL * (X->highexponent - 1);
    if ( s->mantissa <1<<(Dyadic_Base_BITZAHL/2) )
      { s->mantissa = s->mantissa <<(Dyadic_Base_BITZAHL/2);
        s->exponent-=Dyadic_Base_BITZAHL/2;
        if ( X->anzahl > 1) {
          d = ((X->Wert->BlockFeld[BLOCKSIZE - 2])
                         >>(Dyadic_Base_BITZAHL-Dyadic_Base_BITZAHL/2)); 
          s->mantissa+= d;
          if ( (d <<(Dyadic_Base_BITZAHL-Dyadic_Base_BITZAHL/2))
                    !=(X->Wert->BlockFeld[BLOCKSIZE - 2]) )  
                    s->mantissa +=1;
          }
    } else if ( X->anzahl > 1) s->mantissa +=1;
  }
}



long Dyadic_LRSIZE
(const Dyad * X)
{
  unsigned long c, r, m,n;
  c = X->Wert->BlockFeld[BLOCKSIZE - 1];

  if (c == 0) {
    return Dyadic_Base_MinInt;
  } else {
    n=c;
    r=0; 
    for (m=Dyadic_Base_BITZAHL>>1; m>0; m=m>>1 ) {
      if (n>= (1<<m) ){
        r+=m; n=n>>m;
      }
    }
    if (c > (1<<r) || X->anzahl >1) r++;    
    return r + Dyadic_Base_BITZAHL * (X->highexponent - 1);
  }
}


long Dyadic_LRCOMPARE
(const Dyad * X, const Dyad * Y)
{ long sx,sy,lx,ly,erg,i,anz;
  sx=Dyadic_LRSIGN(X);
  sy=Dyadic_LRSIGN(Y);
  if (sx < sy) return -1;
  if (sx > sy) return +1;
  if (sx == 0)  return  0;
  if (sx < 0)  erg=-1; else erg=+1;
  lx=X->highexponent;
  ly=X->highexponent;
  if (lx<ly) return erg;
  if (ly>lx) return -erg;
  LRGET(X, lx, AKKUMULATOR1, 4);
  LRGET(Y, ly, AKKUMULATOR2, 4);
  for (i=0;i<4;i++) {
    if (AKKUMULATOR1[i]<AKKUMULATOR2[i]) return erg;
    if (AKKUMULATOR1[i]>AKKUMULATOR2[i]) return -erg;
  }
  anz=max (X->anzahl,Y->anzahl);
  LRGET(X, lx, AKKUMULATOR1, anz);
  LRGET(Y, ly, AKKUMULATOR2, anz);
  for (i=4;i<anz;i++) {
    if (AKKUMULATOR1[i]<AKKUMULATOR2[i]) return erg;
    if (AKKUMULATOR1[i]>AKKUMULATOR2[i]) return -erg;
  }
  return 0;
}

long Dyadic_LRLENGTH
(const Dyad * X)
{
  return X->anzahl * Dyadic_Base_BITZAHL;
}

void Dyadic_LRCOPY
(const Dyad * X, Dyad* *Z, long N)
{
  long n;

  if (N > 0) {
    n = (N - 1) / Dyadic_Base_BITZAHL;
  } else {
    n = N / Dyadic_Base_BITZAHL;
    n--;
  }
  if (n >= X->highexponent) {
    Dyadic_INTTOLR(0, Z);
  } else {
    LRGET(X, X->highexponent, AKKUMULATOR1, X->highexponent - n + 1);
    LRSET(Z, X->highexponent, AKKUMULATOR1, X->highexponent - n + 1, X->negativ);
  }
}

void Dyadic_LRABS
(const Dyad * X, Dyad* * Z)
{
    LRGET(X, X->highexponent, AKKUMULATOR1, X->anzahl);
    LRSET(Z, X->highexponent, AKKUMULATOR1, X->anzahl, 0);

}



void Dyadic_LRDUPLICATE
(const Dyad * X, Dyad* * Z)
{
    LRGET(X, X->highexponent, AKKUMULATOR1, X->anzahl);
    LRSET(Z, X->highexponent, AKKUMULATOR1, X->anzahl, X->negativ);
}


void Dyadic_LRTRUNC
(const Dyad * X, Dyad* *Z)
{
  if (X->highexponent <= 0) {
    Dyadic_INTTOLR(0, Z);
  } else {
    LRGET(X, X->highexponent, AKKUMULATOR1, X->highexponent);
    LRSET(Z, X->highexponent, AKKUMULATOR1, X->highexponent, X->negativ);
  }
}

void Dyadic_LRFRAC
(const Dyad * X, Dyad* * Z)
{
  long i;

  {

    if (X->highexponent <= 0) {
      LRGET(X, X->highexponent, AKKUMULATOR1, X->anzahl);
      LRSET(Z, X->highexponent, AKKUMULATOR1, X->anzahl, X->negativ);
    } else if (X->highexponent >= X->anzahl) {
      Dyadic_INTTOLR(0, Z);
    } else {
      LRGET(X, X->highexponent, AKKUMULATOR1, X->anzahl);
      {
        long B_53 = 1, B_54 = X->anzahl - X->highexponent;

        if (B_53 <= B_54)
          for (i = B_53;; i += 1) {
            AKKUMULATOR1[i - 1] = AKKUMULATOR1[i + X->highexponent - 1];
            if (i >= B_54) break;
          }
      }
      LRSET(Z, 0, AKKUMULATOR1, X->anzahl - X->highexponent, X->negativ);
    }
  }
}

void Dyadic_LRSPLIT
(const Dyad * X, Dyad* * Y, Dyad* * Z)
{
  LRGET(X, X->highexponent, AKKUMULATOR1, X->anzahl);
  if ( X->highexponent <= 0) {
    LRSET(Z, X->highexponent, AKKUMULATOR1, X->anzahl, X->negativ);
    Dyadic_INTTOLR(0, Y);
  } else if (X->highexponent >= X->anzahl) {
    LRSET(Y, X->highexponent, AKKUMULATOR1, X->anzahl, X->negativ);
    Dyadic_INTTOLR(0, Z);
  } else {
    long highexponent,anzahl,negativ;
    highexponent= X->highexponent;
    anzahl=X->anzahl;
    negativ=X->negativ;
    LRSET(Y, highexponent, AKKUMULATOR1, highexponent, negativ);
    LRSET(Z, 0, &AKKUMULATOR1[highexponent], anzahl - highexponent, negativ);
  }
}

void Dyadic_LRINV
(const Dyad * X, Dyad* * Z, long N)
{
  long n, Anzahl, XAnz;

  if (N - 1 >= 0) {
    n = (N - 1) / Dyadic_Base_BITZAHL;
  } else {
    n = N / Dyadic_Base_BITZAHL;
    n--;
  }
  if (n > -X->highexponent) {
    Dyadic_INTTOLR(0, Z);
  } else {
    AKKUMULATOR1[0] = 1;
    Anzahl = 2 - X->highexponent - n;
    if (Anzahl < X->anzahl) {
      XAnz = Anzahl;
    } else {
      XAnz = X->anzahl;
    }
    LRGET(X, X->highexponent, AKKUMULATOR2, XAnz);
    Dyadic_Base_SysDiv(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, 1, XAnz, Anzahl, FALSE);
    LRSET(Z, 2 - X->highexponent, AKKUMULATOR3, Anzahl, X->negativ);
  }
}

void Dyadic_LRSQRT
(const Dyad * X, Dyad* * Z, long N)
{
  long i, k;
  BOOLEAN LAST;
 
  i = Dyadic_LRSIGN(X);
  if (i < 0) {
    (*Dyadic_LRError)(4);
    Dyadic_INTTOLR(0, Z);
    return;
  }
  if (i == 0) {
    Dyadic_INTTOLR(0, Z);
    return;
  }
  i = Dyadic_LRSIZE(X);
  if (i < 2 * N) {
    Dyadic_INTTOLR(0, Z);
    return;
  }
  i = i / 2;
  Dyadic_INTTOLR(1, &A_LRSQRT);
  Dyadic_LRSHIFT(A_LRSQRT, &A_LRSQRT, i);
  k = i;
  LAST = FALSE;
  for (;;) {
    k = 2 * k - i - 1;
    if (!LAST && k <= (N + i) / 2 - 2) {
      LAST = TRUE;
      k = (N + i) / 2 - 2;
    } else if (k < N) {
      k = N - 1;
    }
    Dyadic_LRDIV(X, A_LRSQRT, &B_LRSQRT, k - 10);
    Dyadic_LRADD(B_LRSQRT, A_LRSQRT, &A_LRSQRT, k - 10);
    Dyadic_LRSHIFT(A_LRSQRT, &A_LRSQRT, -1);
    if (k < N) {
      goto EXIT_4;
    }
  } EXIT_4:;
  Dyadic_LRCOPY(A_LRSQRT, Z, N - 1);
}

void Dyadic_LIADD
(const Dyad * X, const Dyad * Y, Dyad* * Z)
{
  long Anzahl, HighExp;
  BOOLEAN AkkuNegativ;

  if (X->highexponent > Y->highexponent) {
    HighExp = X->highexponent;
  } else {
    HighExp = Y->highexponent;
  }
  if (X->highexponent - X->anzahl > Y->highexponent - Y->anzahl) {
    Anzahl = HighExp - Y->highexponent + Y->anzahl;
  } else {
    Anzahl = HighExp - X->highexponent + X->anzahl;
  }
  LRGET(X, HighExp, AKKUMULATOR1, Anzahl);
  LRGET(Y, HighExp, AKKUMULATOR2, Anzahl);
  if (X->negativ == Y->negativ) {
    Dyadic_Base_SysAdd(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, Anzahl);
    LRSET(Z, HighExp + 1, AKKUMULATOR3, Anzahl + 1, X->negativ);
  } else {
    Dyadic_Base_SysSub(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, Anzahl, &AkkuNegativ);
    if (AkkuNegativ) {
      LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, !X->negativ);
    } else {
      LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, X->negativ);
    }
  }
}

void Dyadic_LISUB
(const Dyad * X, const Dyad * Y, Dyad* *Z)
{
  long Anzahl, HighExp;
  BOOLEAN AkkuNegativ;

  if (X->highexponent > Y->highexponent) {
    HighExp = X->highexponent;
  } else {
    HighExp = Y->highexponent;
  }
  if (X->highexponent - X->anzahl > Y->highexponent - Y->anzahl) {
    Anzahl = HighExp - Y->highexponent + Y->anzahl;
  } else {
    Anzahl = HighExp - X->highexponent + X->anzahl;
  }
  LRGET(X, HighExp, AKKUMULATOR1, Anzahl);
  LRGET(Y, HighExp, AKKUMULATOR2, Anzahl);
  if (X->negativ != Y->negativ) {
    Dyadic_Base_SysAdd(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, Anzahl);
    LRSET(Z, HighExp + 1, AKKUMULATOR3, Anzahl + 1, X->negativ);
  } else {
    Dyadic_Base_SysSub(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, Anzahl, &AkkuNegativ);
    if (AkkuNegativ) {
      LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, !X->negativ);
    } else {
      LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, X->negativ);
    }
  }
}

void Dyadic_LIMULT
(const Dyad * X, const Dyad * Y, Dyad* *Z)
{
  long HighExp, Anzahl;

  HighExp = X->highexponent + Y->highexponent;
  Anzahl = X->anzahl + Y->anzahl;
  LRGET(X, X->highexponent, AKKUMULATOR1, X->anzahl);
  LRGET(Y, Y->highexponent, AKKUMULATOR2, Y->anzahl);
  Dyadic_Base_SysMul(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, X->anzahl, Y->anzahl, Anzahl);
  LRSET(Z, HighExp, AKKUMULATOR3, Anzahl, X->negativ != Y->negativ);
}

static void SPLIT (const Dyad * X, Dyad* * Y, Dyad* * Z, long s)
{
  LRGET(X, X->highexponent, AKKUMULATOR1, X->anzahl);
  if (X->highexponent - X->anzahl >= s) {
    LRSET(Y, X->highexponent - s, AKKUMULATOR1, X->anzahl, X->negativ);
    Dyadic_INTTOLR(0, Z);
  } else { 
    long highexponent,anzahl,negativ;
    highexponent= X->highexponent;
    anzahl=X->anzahl;
    negativ=X->negativ;
    LRSET(Y, highexponent - s, AKKUMULATOR1, highexponent - s, negativ);
    { long i;
      LONGINT B_54 = anzahl + s - highexponent;
      for (i = 0;i<B_54; i += 1) 
        AKKUMULATOR1[i] = AKKUMULATOR1[i + highexponent - s];
    }
    LRSET(Z, s, AKKUMULATOR1, anzahl + s - highexponent, negativ);
  }
}

void Dyadic_LIMULTKARAT (const Dyad * X, const Dyad * Y, Dyad* * Z)
{

  Dyad *A, *B, *C, *D, *P;
  long ix, iy, imax, imin, splitwert;

  ix = X->highexponent;
  iy = Y->highexponent;
  if (ix < iy) {
    imin = ix;
    imax = iy;
  } else {
    imin = iy;
    imax = ix;
  }
  if (imin < trigger) {
    Dyadic_LIMULT(X, Y, Z);
    return;
  }
  splitwert = imax / 2;
  Dyadic_LRINIT(&A);
  Dyadic_LRINIT(&B);
  Dyadic_LRINIT(&C);
  Dyadic_LRINIT(&D);
  Dyadic_LRINIT(&P);
  SPLIT(X, &A, &B, splitwert);
  SPLIT(Y, &C, &D, splitwert);
  Dyadic_LIADD(A, B, &P);
  Dyadic_LIADD(C, D, Z);
  Dyadic_LIMULTKARAT(P, *Z, Z);
  Dyadic_LIMULTKARAT(B, D, &B);
  Dyadic_LIMULTKARAT(A, C, &A);
  Dyadic_LISUB(*Z, B, Z);
  Dyadic_LISUB(*Z, A, Z);
  INC1((*Z)->highexponent, splitwert);
  Dyadic_LIADD(*Z, B, Z);
  INC1(A->highexponent, 2 * splitwert);
  Dyadic_LIADD(A, *Z, Z);
  Dyadic_LRCLEAR(&A);
  Dyadic_LRCLEAR(&B);
  Dyadic_LRCLEAR(&C);
  Dyadic_LRCLEAR(&D);
  Dyadic_LRCLEAR(&P);
}



void Dyadic_LIDIV
(const Dyad * X, const Dyad * Y, Dyad* * Q, Dyad* * R)
{
  long Anzahl;
  BOOLEAN Neg;

  if (Dyadic_LRSIGN(Y) == 0) {
    Dyadic_INTTOLR(0, Q);
    Dyadic_INTTOLR(0, R);
    (*Dyadic_LRError)(2);
    return;
  }
  Anzahl = X->highexponent - Y->highexponent + 1;
  if (Anzahl <= 0) {
    LRGET(X, X->highexponent, AKKUMULATOR1, X->anzahl);
    LRSET(R, X->highexponent, AKKUMULATOR1, X->anzahl, X->negativ);
    Dyadic_INTTOLR(0, Q);
  } else {
    LRGET(X, X->highexponent, AKKUMULATOR1, X->anzahl);
    LRGET(Y, Y->highexponent, AKKUMULATOR2, Y->highexponent);
    Dyadic_Base_SysDiv(AKKUMULATOR1, AKKUMULATOR2, AKKUMULATOR3, X->anzahl, Y->highexponent, Anzahl, TRUE);
    Neg = X->negativ != Y->negativ;
    LRSET(Q, Anzahl, AKKUMULATOR3, Anzahl, Neg);
    LRSET(R, Y->highexponent, AKKUMULATOR1, Y->highexponent, Neg);
  }
}

static void LICOPY
(const Dyad * X, Dyad* * Z)
{
  Dyadic_LRDUPLICATE(X,Z);
}

void BEGIN_Dyadic()
{
  static BOOLEAN has_been_called = FALSE;

  if (!has_been_called) {
    has_been_called = TRUE;

    Initialize();
  }
}
