#include "MPFR_ext.h"
#define BITS_PER_MP_LIMB 32
                                      
#define   min(a,b)        (a < b ? a : b)
#define   max(a,b)        (a < b ? b : a)
#define   max3(a,b,c)       max(a,max(b,c))
#define   max4(a,b,c,d)       max(max(a,b),max(c,d))


long ext_mpfr_var_count=0;

static ext_mpfr_type mpfr_FreeVars=0L;
static long mpfr_FreeVarCount=0L;

void ext_mpfr_initialize()
{
mpfr_set_default_prec(32);
}

ext_mpfr_type ext_mpfr_init()
{ ext_mpfr_type z;
  if (mpfr_FreeVars) {
    z =mpfr_FreeVars;
    mpfr_FreeVars=mpfr_FreeVars->_mp_exp;
    mpfr_FreeVarCount -=1;
  } else {
  z = (ext_mpfr_type) malloc(sizeof (__mpfr_struct));
  mpfr_init(z);
  }
  ext_mpfr_var_count +=1;
  return z;
}

void ext_mpfr_free(ext_mpfr_type z)
{
  if ( mpfr_FreeVarCount < 1000 ) {
    mpfr_FreeVarCount +=1;
    z->_mp_exp=mpfr_FreeVars;
    mpfr_FreeVars=z;
  } else {
    mpfr_clear(z); 
    free (z);
  }
  ext_mpfr_var_count -=1;
}

long ext_mpfr_size(ext_mpfr_type z)
{
if NOTZERO(z) return EXP(z);
else return GMP_min;
}

void ext_mpfr_getsize(ext_mpfr_type z,ext_mpfr_sizetype* s)
{
  if ( ABSSIZE(z)==0 ) {
    s->mantissa=0;
    s->exponent=GMP_min;
  } else if ( ((z->_mp_d)[ABSSIZE(z)-1]==0) ) {
    s->mantissa=0;
    s->exponent=GMP_min;
  } else {
    s->mantissa=(z->_mp_d)[ABSSIZE(z)-1];
    s->exponent=EXP(z)-BITS_PER_MP_LIMB;
    if (ABSSIZE(z)>0&& abs(z->_mp_size+1)>0) s->mantissa+=1;
  }
}

void ext_mpfr_add(ext_mpfr_type z1,ext_mpfr_type z2,ext_mpfr_type z,long p)
{ long q,s1,s2;
  s1=ext_mpfr_size(z1);
  s2=ext_mpfr_size(z2);
  q=max(s1,s2)-p+10;
  if (q<10)q=10;
  mpfr_set_prec(z,q);
  mpfr_add(z,z1,z2,GMP_RNDZ);
  return;
}

void ext_mpfr_sub(ext_mpfr_type z1,ext_mpfr_type z2,ext_mpfr_type z,long p)
{ long q,s1,s2;
  s1=ext_mpfr_size(z1);
  s2=ext_mpfr_size(z2);
  q=max(s1,s2)-p+10;
  if (q<10)q=10;
  mpfr_set_prec(z,q);
  mpfr_sub(z,z1,z2,GMP_RNDZ);
  return;
}

void ext_mpfr_mul(ext_mpfr_type z1,ext_mpfr_type z2,ext_mpfr_type z,long p)
{ long q,s1,s2;
  s1=ext_mpfr_size(z1);
  s2=ext_mpfr_size(z2);
  q=s1+s2-p+40;
  if (q<10)q=10;
  mpfr_set_prec(z,q);
  mpfr_mul(z,z1,z2,GMP_RNDZ);
  return;
}

void ext_mpfr_div(ext_mpfr_type z1,ext_mpfr_type z2,ext_mpfr_type z,long p)
{ long q,s1,s2;
  s1=ext_mpfr_size(z1);
  s2=ext_mpfr_size(z2);
  q=s1-s2-p+10;
  if (q<10)q=10;
  mpfr_set_prec(z,q);
  mpfr_div(z,z1,z2,GMP_RNDZ);
  return;
}

void ext_mpfr_abs(ext_mpfr_type z1,ext_mpfr_type z)
{
  long q1=mpfr_get_prec(z1);
  mpfr_set_prec(z,q1);
  mpfr_set(z,z1,GMP_RNDZ);
  CHANGE_SIGN(z);
  return;
}

void ext_mpfr_truncate(ext_mpfr_type z1,ext_mpfr_type z)
{
  long q1=mpfr_get_prec(z1);
  mpfr_set_prec(z,q1-EXP(z));
  mpfr_set(z,z1,GMP_RNDZ);
  return;
}


void ext_mpfr_writee(ext_mpfr_type z,long w)
{ char s[w];
  long e; 
  int i;
  int flag=0;
  mpfr_get_str (s, &e, 10, w-7,z,GMP_RNDZ);
  for (i=0;i<w-7;i++) {
    if ( s[i]==0 ) flag=1 ;
    if ( flag ) s[i]='0';
  } 
  if (s[0]=='-') {s[w-7]=0;printf("-.%sE%+05ld",&s[1],e);}
  else {s[w-8]=0;printf("+.%sE%+05ld",&s[0],e);};
}

void ext_mpfr_write(ext_mpfr_type z,long w)
{ char s[w];
  long e; 
  int i;
  int flag=0;
  mpfr_get_str (s, &e, 10, w-7,z,GMP_RNDZ);
  for (i=0;i<w-7;i++) { if ( s[i]==0 ) flag=1 ;
    if ( flag ) s[i]='0';
  } 
  if (s[0]=='-') {s[w-7]=0;printf("-.%sE%+05ld",&s[1],e);}
  else {s[w-8]=0;printf("+.%sE%+05ld",&s[0],e);};
  if (z->_mp_size != 0) {
     printf("\n");
     for (i=abs(z->_mp_size-1);i>=0; i-=1) 
       printf("%lu ",z->_mp_d[i]);
  }
  printf("\nExponent %ld, Limbs: %d, Bits: %ld\n",
     z->_mp_exp,z->_mp_size,mpfr_get_prec(z));
}

void ext_mpfr_duplicate_wo_init(ext_mpfr_type z1,ext_mpfr_type z2)
{
  long q1=mpfr_get_prec(z1);
  if (mpfr_get_prec(z2)< q1) mpfr_set_prec(z2,q1);
  mpfr_set(z2,z1,GMP_RNDZ);
}

void ext_mpfr_duplicate_w_init(ext_mpfr_type z1,ext_mpfr_type *z2)
{
//  *z2 = (ext_mpfr_type) malloc(sizeof (__mpfr_struct));
//  mpfr_init2(*z2,mpfr_get_prec(z1));
  long q1=mpfr_get_prec(z1);
  *z2=ext_mpfr_init();
  if (mpfr_get_prec(*z2)< q1) mpfr_set_prec(*z2,q1);
  ext_mpfr_var_count +=1;
  mpfr_set(*z2,z1,GMP_RNDZ);
}

void ext_mpfr_sqrt(ext_mpfr_type z1,ext_mpfr_type z,long p)
{ long q,s1,s2;
  s1=ext_mpfr_size(z1);
  q=max(s1,s2)-p+10;
  if (q<10)q=10;
  mpfr_set_prec(z,q);
  mpfr_sqrt(z,z1,GMP_RNDZ);
  return;
}

void ext_mpfr_shift(ext_mpfr_type z1,ext_mpfr_type z,long n)
{
  mpfr_set_prec(z,mpfr_get_prec(z1));
  if ( n>=  0 ) mpfr_mul_2exp(z,z1,n,GMP_RNDZ);
  else mpfr_div_2exp(z,z1,-n,GMP_RNDZ);
  return;
}

