extern "C" {

#include "MPFR_ext.h"

#define MP_type ext_mpfr_type

#define MP_initialize      ext_mpfr_initialize()

#define MP_min    GMP_min
#define MP_max    GMP_max

#define MP_init(z)  z=ext_mpfr_init()

#define MP_clear(z)  ext_mpfr_free(z) 

#define MP_int_to_mp(i,z) mpfr_set_si(z,i,GMP_RNDZ)

// duplicate value z1 to z2, with/without initialization of z2 
#define MP_duplicate_w_init(z1,z2)  ext_mpfr_duplicate_w_init(z1,&(z2))
#define MP_duplicate_wo_init(z1,z2) ext_mpfr_duplicate_wo_init(z1,z2) 

// copy z1 to z2, but precision p is sufficient
#define MP_copy(z1,z2,p) ext_mpfr_duplicate_wo_init(z1,z2)

#define MP_add(z1,z2,z,p) ext_mpfr_add(z1,z2,z,p)
#define MP_sub(z1,z2,z,p) ext_mpfr_sub(z1,z2,z,p)
#define MP_mul(z1,z2,z,p) ext_mpfr_mul(z1,z2,z,p) 
#define MP_div(z1,z2,z,p) ext_mpfr_div(z1,z2,z,p) 
#define MP_abs(z1,z) ext_mpfr_abs(z1,z)

// MP_compare: values -1 (lessthan), 0 (equal), or 1 (greater than)
#define MP_compare(z1,z2) mpfr_cmp(z1,z2)

// MP_sign: values -1 (negative), 0 (zero), or 1 (positive)
#define MP_sign(z) SIGN(z)

// long MP_size : smallest(?) value k such that 2**k > z
#define MP_size(z) ext_mpfr_size(z)
#define MP_getsize(z,s) ext_mpfr_getsize(z,(ext_mpfr_sizetype*)&s)

// truncate z1 yielding z2 
#define MP_truncate(z1,z) ext_mpfr_truncate(z1,z)

// In/Out
#define MP_writee(z,w) ext_mpfr_writee(z,w)
#define MP_writef(z,w,d) mpfr_out_str(stdout,10,w,z,GMP_RNDZ)
#define MP_write(z,c) ext_mpfr_write(z,c)

// Statistics

#define MP_var_count ext_mpfr_var_count
#define MP_space_count long(111111111)
#define MP_max_space_count  long(111111111)

// special functions
#define MP_has_sqrt 1
#define MP_sqrt(z1,z,p) ext_mpfr_sqrt(z1,z,p)

#define MP_has_shift 1
#define MP_shift(z1,z,n) ext_mpfr_shift(z1,z,n)
}
