#include "COMPLEX.h"

#include "REALlimits.h"

COMPLEX::COMPLEX():REALMATRIX(2,1)
{}

COMPLEX::COMPLEX(const REAL& x):REALMATRIX(2,1)
{
  element(0,0)=x;
}

COMPLEX::COMPLEX(const REAL& x, const REAL& y):REALMATRIX(2,1)
{
  element(0,0)=x;
  element(1,0)=y;
}

COMPLEX::COMPLEX(const REALMATRIX& x):REALMATRIX(2,1)
{
  element(0,0)=x(0,0);
  element(1,0)=x(1,0);
}

COMPLEX operator + (const COMPLEX& x, const COMPLEX& y)
{
  return COMPLEX(x(0,0)+y(0,0),
                 x(1,0)+y(1,0));
}

COMPLEX operator - (const COMPLEX& x, const COMPLEX& y)
{
  return COMPLEX(x(0,0)-y(0,0),
                 x(1,0)-y(1,0));
}

COMPLEX operator - (const COMPLEX& x)
{
  return COMPLEX(-x(0,0),
                 -x(1,0));
}

COMPLEX operator * (const COMPLEX& x, const COMPLEX& y)
{
  return COMPLEX(x(0,0)*y(0,0)-
                 x(1,0)*y(1,0),
                 x(0,0)*y(1,0)+
                 x(1,0)*y(0,0));
}

COMPLEX operator / (const COMPLEX& x, const COMPLEX& y)
{ 
  REAL d;
  d=y(0,0)*y(0,0) +y(1,0)*y(1,0);
  return COMPLEX( (x(0,0)*y(0,0)+
                   x(1,0)*y(1,0))/d,
                  (x(1,0)*y(0,0)-
                   x(0,0)*y(1,0))/d);
}

REAL real(const COMPLEX& z)
{ return z(0,0); }

REAL imag(const COMPLEX& z)
{ return z(1,0); }

REAL abs(const COMPLEX& z)
{ return sqrt(z(0,0)*z(0,0)+z(1,0)*z(1,0)); }


COMPLEX c_sqrt_approx(long p, long* choice, const COMPLEX& z)
{
  REAL a,b,c,d,r;
  COMPLEX y;
  r= abs (z);
  a= real(z);
  b= imag(z);
  c= sqrt( (r+a) / 2L );
  d= sqrt( (r-a) / 2L );
  if ( *choice == 0 ) {
    if ( !bound(r, 2*p) )
      if ( ! bound(b,p-10) ) {
        if ( b > 0 )  *choice=1; 
                 else  *choice=2; 
      } else {
        if ( a > 0 )  *choice=3; 
                 else  *choice=4; 
      }
    }

  switch ( *choice ) {
    case 0:    y = COMPLEX (  0     , 0     );    break;
    case 1:    y = COMPLEX (  c      ,  d     );    break;
    case 2:    y = COMPLEX (  c      , -d     );    break;
    case 3:    y = COMPLEX (  c      , b/2L/c );    break;
    case 4:    y = COMPLEX (  b/2L/d ,  d     );    break;
  }
  return y;
}

COMPLEX  sqrt(const COMPLEX& z)
{
  return limit_mv(c_sqrt_approx,z);
}

COMPLEX csqrt(const COMPLEX& z)
{
  REAL a,b,c,d,r,s;
  COMPLEX y;
  long choice=0;
 
  r= abs (z);
  a= real(z);
  b= imag(z);
  c= sqrt( (r+a) / 2 );
  d= sqrt( (r-a) / 2 );

  if ( ! bound(b,size(r)-3) ) {
    if ( b > 0 )  choice=1;  else  choice=2; 
  } else {
    if ( a > 0 )  choice=3;  else  choice=4; 
  }

  switch ( choice ) {
  case 1:    y = COMPLEX ( c      ,   d    ); break;
  case 2:    y = COMPLEX ( c      ,  -d    ); break;
  case 3:    y = COMPLEX ( -c     , -b/2/c ); break;
  case 4:    y = COMPLEX ( b/2/d ,   d    ); break;
  }
  return y;
}
