#include "stdio.h"
#include "stdlib.h"
#include "REALLIB.h"
#include "REALlimits.h"
#include "REALcore.h"

REAL strtoREAL(char* s, char** endptr){
  stiff_begin();
  int exp=0;
  int sign=1;
  REAL y;
  REAL ten=10;
  *endptr=s;

  while ( **endptr == '0' ) *endptr+=1;

  if ( **endptr == '-' ) {sign=-1;*endptr+=1;}
  else if ( **endptr == '+' ) {sign=1;*endptr+=1;}

  while ( **endptr >= '0' &&  **endptr <= '9' ) {
  y=ten*y + REAL(sign*(**endptr-'0')); *endptr+=1;}

  if ( **endptr == '.' ) *endptr+=1;

  while ( **endptr >= '0' &&  **endptr <= '9' ) {
  y=ten*y + REAL(sign*(**endptr-'0')); exp-=1;*endptr+=1;}

  if ( **endptr == 'E' || **endptr == 'e'  ) {
  *endptr+=1;exp+=strtol(*endptr,endptr,10); }

  if (exp !=0) y=y*power(ten,exp);

  stiff_end();
  return y;
}

REAL atoREAL(char* s){
  char*dummy;
  return strtoREAL(s,&dummy);
};


REAL modulo (const REAL& x, const REAL& y){
   return x-round2(x/y)*y;
};


REAL power(const REAL& x, long n) {
   if (n==0) return 1;
   if (n==1) return x;
   stiff_begin();
   REAL y=1;
   REAL xc=x;
   if (n<0) {xc=y/x;n=-n;}
   if (n==1) {
     stiff_end();
     return xc;
   }
   for (long k=n;k>0;k=k/2) { 
     if (k%2==1) y=y*xc;
     if ( k ==1) break;
     xc=xc*xc;}
   stiff_end();
   return y; 
}

REAL root_approx(long prec, const REAL& x,long n) {
   long u=upperbound(x)/n;
   REAL error;
   if (u < prec) return 0;
   REAL r=scale(REAL(1),u);
   REAL r_2=x/power(r,n-1);
   do {
      r=r+(r_2-r)/n;
      r_2=x/power(r,n-1);
      error=r - r_2;
   } while ( !bound(error,prec) );
   return r;
};

REAL root(const REAL& x,long n) {
   return limit(root_approx,x,n);
}

REAL max_approx (long prec, const REAL& x, const REAL& y){
   if ( positive(x-y,prec) ) return x;
   else return y;
};

REAL maximum (const REAL& x, const REAL& y){
   return limit_lip(max_approx,0,x,y);
};

REAL min_approx (long prec, const REAL& x, const REAL& y){
   if ( positive(x-y,prec) ) return y;
   else return x;
};

REAL minimum (const REAL& x, const REAL& y){
   return limit_lip(min_approx,0,x,y);
};

/*****************************************************************/
/*                     ln(2)=0.6931471805....                    */
/*   Usage: ln2()                                                */
/*   Application: range reduction, AGM methods for logarithms, ..*/ 
/*****************************************************************/

REAL ln2_val=1;
long ln2_err=0;


REAL ln2_approx (long prec){
  long N=100-prec/2;
  REAL a=1,b=scale(REAL(1),N),c;
  do {
    c=a*b;
    a=2*c/(a+b);
    b=sqrt(c);
//  } while ( !bound(a-b,-N) );
//  c= a*b/(a+b)/(N+2);
  } while ( !bound(a-b,-2*N) );
  c= a/(2*N+4);
  return c;
};

REAL ln2(){
   if (ln2_err >  REAL_actual_prec || ln2_err == 0) {
     ln2_err =  REAL_actual_prec;
     ln2_val = pi()*limit(ln2_approx);
    }
   return ln2_val;
};


/*****************************************************************/
/*                     pi = 3.141592653...                       */
/*   Usage: pi()                                                 */
/*   Application: range reduction, ...                           */ 
/*****************************************************************/


REAL pi_val=3;
long pi_err=0;

REAL pi_approx_MACHIN (long prec){
   REAL z1=REAL(1)/REAL(5);
   REAL z2=REAL(1)/REAL(239);
   REAL y,z;
   long i;
  
   i=3;
   y=z1;
   z=y;
   while ( !bound(z,prec-5) ) {
     y=-y/(5*5);
     z=y/i;
     z1=z1+z;
     i+=2;
   }
   i=3;
   y=z2;
   z=y;
   while ( !bound(z,prec-3) ) {
     y=-y/(239*239);
     z=y/i;
     z2=z2+z;
     i+=2;
   }
   return 4*(4*z1-z2);
};

REAL pi_approx_AGM (long prec){
   REAL a= 1;
   REAL b= 1/sqrt(REAL(2));
   REAL t= 1/REAL(4);
   REAL pi_l,pi_u;
   long j=1;
   REAL y;
   do {
     y=a;
     a=(a+b)/2;
     b=sqrt(b*y);
     t=t-j*power(a-y,2);
     j=2*j; 
   } while ( ! bound(a-b, prec));
   return a*a/t;
};

REAL pi_inv_approx_BORWEIN (long prec){
   REAL a= 6-4*sqrt(REAL(2));
   REAL y= sqrt(REAL(2))-1;
   REAL s=2,t,aold;
   long k=1;
   do {
     aold=a;
     t=sqrt(sqrt(1-power(y,4)));
     y=(1-t)/(1+t);
     s=s*4;
     a=a*power(1+y,4)-s*y*(1+y+y*y);
     k=2*k; 
   } while (  !bound (a-aold,prec/4 )) ;
   return a;
};

REAL pi (){
   if (pi_err >  REAL_actual_prec || pi_err == 0) {
     pi_err =  REAL_actual_prec;
//   pi_val = limit(pi_approx_MACHIN);
//   pi_val = limit(pi_approx_AGM);
     pi_val = 1/limit(pi_inv_approx_BORWEIN);
    }
   return pi_val;
};

/*****************************************************************/
/*                     sin(x), cos(x), tan(x)                    */
/*****************************************************************/


REAL sin_approx (long prec, const REAL& x){
   if ( prec >= 0 ) return 0;
   REAL z=x,y=x,x2=x*x;
   long i=2;
   while ( !bound(y,prec) || ! bound(y,-1) ) {
     y=-y*x2/(i*(i+1));
     z=z+y;
     i+=2;
   }
   return z;
};

REAL cos_approx (long prec, const REAL& x){
   if ( prec >= 0 ) return 0;
   REAL z=1,y=z,x2=x*x;
   long i=1;
   while ( !bound(y,prec) || !bound(y,-1) ) {
     y=-y*x2/(i*(i+1));
     z=z+y;
     i+=2;
   }
   return z;
};

REAL cos3 (const REAL& x){
  REAL y=limit_lip(cos_approx,0,x/59049);
  for (long i=1;i<=10;i+=1) y=(4*y*y-3)*y;
  return y;
};

REAL sin3 (const REAL& x){
  REAL y=limit_lip(sin_approx,0,x/59049);
  for (long i=1;i<=10;i+=1) y=y*(3-4*y*y);
  return y;
};


REAL cos (const REAL& x){
  continous_begin();
  REAL m=modulo(x,2*pi());
  long i=round(m*100);
  if ( i <  0 ) {i=i+314;m=m+2*pi();}
  if ( i> 550 )
    m= cos3(m-2*pi());
  else if ( i> 392)
    m= sin3(m-3*pi()/2);
  else if ( i> 235)
    m= -cos3(pi()-m);
  else if ( i> 80)
    m= sin3(pi()/2-m);
  else
    m= cos3(m);
  continous_end();
  return m;
};

REAL sin (const REAL& x){
  continous_begin();
  REAL m=modulo(x,2*pi());
  long i=round(m*100);
  if ( i <  0 ) {i=i+314;m=m+2*pi();}
  if ( i> 550 )
    m= sin3(m-2*pi());
  else if ( i> 392)
    m= -cos3(m-3*pi()/2);
  else if ( i> 235)
    m= sin3(pi()-m);
  else if ( i> 80)
    m= cos3(pi()/2-m);
  else
    m= sin3(m);
  continous_end();
  return m;
};


REAL tan (const REAL& x){
   return sin(x)/cos(x);
};

///////////////////////////////////////////////////////////////////

REAL atan_approx (long prec, const REAL& x){
   if ( prec >= 2 ) {int dummy=0; return REAL(dummy);}
   long i=2;
   REAL x2=x*x,y=-x*x2,z=x+y/3;
   while ( !bound(y,prec) || ! bound(y,-1) ) {
     y=-y*x2;
     z=z+y/(i+3);
     i+=2;
   }
   return z;
};

REAL atan (const REAL& x){
  if ( positive(0.01-abs(x),-10) )
    return limit_lip(atan_approx,2,x);
  if ( positive(2-abs(x),-10) ) { 
    REAL y= (sqrt(x*x+1)-1)/x;
    y=  (sqrt(y*y+1)-1)/y;
    return 8*limit_lip(atan_approx,2,(sqrt(y*y+1)-1)/y);
    }
  if ( x > 0   ) return   pi()/2 - atan(1/x);
  else           return -(pi()/2 - atan(-1/x));

};

REAL asin (const REAL& x){
   return atan(x/sqrt(1-x*x));
};

REAL acos (const REAL& x){
   return pi()/2 - asin(x);
};

REAL asec (const REAL& x){
   return acos(1/x);
};

REAL acosec (const REAL& x){
   return asin(1/x);
};

///////////////////////////////////////////////////////////////////
REAL sec (const REAL& x) {
 return 1/cos(x);
}

REAL cosec (const REAL& x) {
 return 1/sin(x);
}

REAL cotan (const REAL& x) {
 return cos(x)/sin(x);
}

REAL sech (const REAL& x) {
 return 2/(exp(x)+exp(-x));
}

REAL cosech (const REAL& x) {
 return 2/(exp(x)-exp(-x));
}

REAL sinh (const REAL& x) {
 return (exp(x)-exp(-x))/2;
}

REAL cosh (const REAL& x) {
 return (exp(x)+exp(-x))/2;
}

REAL tanh (const REAL& x) {
 return (exp(x)-exp(-x))/(exp(x)+exp(-x));
}

REAL coth (const REAL& x) {
 return (exp(x)+exp(-x))/(exp(x)-exp(-x));
}

REAL asinh (const REAL& x) {
 return log(x+sqrt(x*x+1));
}

REAL acosh (const REAL& x) {
 return log(x+sqrt(x*x-1));
}

REAL atanh (const REAL& x) {
 return log((1+x)/(1-x))/2;
}

REAL acoth (const REAL& x) {
 return log((1-x)/(1+x))/2;
}


REAL asech (const REAL& x){
   return acosh(1/x);
};

REAL acosech (const REAL& x){
   return asinh(1/x);
};




///////////////////////////////////////////////////////////////////

REAL exp_approx (long prec, const REAL& x){
  if ( prec >= 2 ) return REAL(long(0));
  REAL z=1,y=z;
  long i=1;
  while ( !bound(y,prec) ) {
    y=y*x/i;
    z=z+y;
    i+=1;
  }
  precision_policy(ABSOLUTE);
  return z;
};

REAL exp_lip (long prec,const REAL& x)
{
  REAL y;
  if ( bound(x,-10) ) {
    y = exp_approx(prec,x);
  }
  else   
    {
    y= exp_lip(prec,x/64);
    y= power(y,64);
    }
  return y;
};

REAL exp (const REAL& x)
{
  continous_begin ();
  REAL y;
  if ( bound(x,-5) )
    y = limit_lip(exp_approx,-7,x);
  else   
    {
    long i=(round(x)+1)*1.443;
    y= limit_lip(exp_lip,i,x);
    }
  continous_end ();
  return y;
};

REAL euler_approx (long prec)
{
  if ( prec >= 2 ) return 0;
  REAL z=1,y=z;
  long i=1;
  while ( !bound(y,prec) ) {
    y=y/i;
    z=z+y;
    i+=1;
  }
  return z;
};

REAL euler ()
{
  return limit(euler_approx);
};


//////////////////////////////////////////////////////////////////


#ifdef MP_has_sqrt

  REAL MP_sqrt_approx (long prec, const REAL& x){
    DYADIC xd=approx(x,prec-1);
    MP_sqrt(xd.value,xd.value,prec-1);
    return REAL(xd);
  };
  #define SQRT_APPROX MP_sqrt_approx

#else

  REAL gen_sqrt_approx(long k,const REAL& x){
    REAL a=1,b=x/a;
    do {
       a=(a+b)/2;
       b=x/a;
    } while ( !bound(a-b,k) );
    return a;
  }
  #define SQRT_APPROX gen_sqrt_approx

#endif

REAL sqrt_approx (long prec, const REAL& x){
  if (bound(x, 2*prec) ) return 0;
  long s2=size(x)/2;
  REAL y=scale(x,-s2*2);
  y=lipschitz(SQRT_APPROX,2,prec-s2,y);
  y=scale(y,s2);
  return y; 
};


REAL sqrt (const REAL& x){
 return limit(sqrt_approx,x);
};

///////////////////////////////////////////////////////////////////

REAL log_taylor_approx(long prec, const REAL& x0) {
// Computing series expansion for log(x) for x > 1/2
    REAL x=sqrt(x0);
    for (long i=1;i<=9;i+=1) x=sqrt(x);
   REAL xm1=x-1;
   REAL y=xm1/x;
   REAL f=y;
   REAL z=y;
   REAL s=y;
   long i=2;
   while ( !bound(s,prec) ) {
     y=y*f;
     s=y/i;
     z=z+s;
     i+=1;
   }
   return z*1024;
;
}

REAL inv_2agm_approx(long prec, const REAL& x, const REAL& y) {
// Computing 2*AGM(1,1/x)
  REAL a=x,b=y,c;
  do {
    c=a*b;
    a=2*c/(a+b); 
    b=sqrt(c);
  } while ( !bound(a-b,prec/2-100) );
  return a*b/(a+b);
//  } while ( !bound(a-b,prec) );
//    return a;
}


REAL agm_approx(long prec, const REAL& x, const REAL& y) {
// Computing 2*AGM(x,y)
  REAL a=x,b=y,c;
  do {
    c=a*b;
    a=scale(a+b,-1); 
    b=sqrt(c);
  } while ( !bound(a-b,prec/2) );
  return a+b;
//  } while ( !bound(a-b,prec) );
//  return 2*a;
}

REAL log_agm_approx(long prec, const REAL& z) {
  long N=100-prec/2;
//  REAL delta= scale(REAL(1),N);
//  REAL l= ln2()*(N+2)-pi()*delta/agm_approx(prec-2-50,delta,z);
  REAL delta= scale(REAL(1),N);
  REAL l= ln2()*(N+2)-pi()*inv_2agm_approx(prec-2,1,delta/z);
  return l;
}

REAL log_approx(long prec, const REAL& z) {
  if ( prec > -1000) return log_taylor_approx(prec,z);
  return log_agm_approx(prec,z);
}


REAL log(const REAL& x) {
 long s=size(x);
 REAL y= sqrt(sqrt(scale(x,2-s)));
 y=limit_lip(log_approx,2,y);
 return 4*y+(s-2)*ln2();
}


/*****************************************************************/
/*                     matrix arithmetic                         */
/*****************************************************************/




REALMATRIX expmatrix_approx (long prec, const REALMATRIX& x){
   if (rows(x)!=columns(x)) {
     fprintf(stderr,"Exponentiation of non-quadratic real matrix [%d,%d]\n",
           columns(x),rows(x));
     exit(1);
     }
   if ( prec >= 2 ) return eye(rows(x));
   REALMATRIX y=x,z=y+eye(rows(x));
   long i=2;
   while ( !bound(y,prec) ) {
     y=y*x/i;
     z=z+y;
     i+=1;
   }
   return z;
};

REALMATRIX exp (const REALMATRIX& x){
continous_begin ();
  REALMATRIX y;
  if ( bound(x,-2) )
  y = limit_lip(expmatrix_approx,0,x);
  else
    {
    y= exp(x/2);
    y= y*y;
    }
continous_end ();
return y;
}

REALMATRIX  steady_state (const REALMATRIX& x) {
  continous_begin ();
  REALMATRIX result(1,rows(x));
  REALMATRIX chain=x;
  REAL factor,outrate;
  if ( rows(x) != columns(x)) { 
     fprintf(stderr,"Argument is not a square matrix\n");
     exit(1);
     }
  for (unsigned int N=0; N < rows(x); N++) chain(N,N)=chain(N,N)-1;

  for (unsigned int N=0; N < rows(x)-1; N++) {
    for (unsigned int i=N+1; i < columns(x); i++) {
      factor=chain(i,N)/chain(N,N);
      for (unsigned int j=N+1; j< columns(x); j++) 
          chain(i,j)=chain(i,j)
                   -factor*chain(N,j);
  } }

  result(0,rows(x)-1)=1;
  for (int N=rows(x)-2; N >= 0; N--) {
   outrate=0;
   for (int j=rows(x)-1; j>N;j--) outrate=outrate
                 +result(0,j)*chain(j,N);
   result(0,N)=outrate/(outrate-chain(N,N));
   factor=-chain(N,N)/(outrate-chain(N,N));
   for (int j=rows(x)-1; j>N;j--)
      result(0,j)=result(0,j)*factor;
  } 
  continous_end ();
  return result;
}


REALMATRIX  steady_state (const SPARSEREALMATRIX& x) {
//  continous_begin ();
rprintf("Teil 1\n");
  REALMATRIX result(1,rows(x));
  SPARSEREALMATRIX chain=x;
  REAL factor,outrate;
  SPM_ELEMENT* a;
  SPM_ELEMENT* b;
  if ( rows(x) != columns(x)) { 
     fprintf(stderr,"Argument is not a square matrix\n");
     exit(1);
     }
  for (unsigned int N=0; N < rows(x); N++)
    sparse_set(chain,N,N,chain(N,N)-1);

rprintf("\nTeil 2, Fuellung jetzt: %d\n",chain.filled);

  for (unsigned int N=0; N < chain.maxrow-1; N++) {
    a=chain.colvector[N];
    while ( a != NULL ) {
      if (a->rowindex >N) { 
        factor=a->value/chain(N,N);
        b=chain.rowvector[N];
        while ( b != NULL ) {
          if (b->colindex >N) 
           sparse_set(chain,a->rowindex,b->colindex,
              chain(a->rowindex,b->colindex)-factor* b->value );
          b=b->nextcol;
        }
      }
      a=a->nextrow;
  } }

rprintf("\nTeil 3, Fuellung jetzt: %d\n",chain.filled);
  result(0,rows(x)-1)=1;
  for (int N=rows(x)-2; N >= 0; N--) {
   outrate=0;

   a=chain.colvector[N];
   while ( a!= NULL  ) {
     if (a->rowindex > N) outrate=outrate+ result(0,a->rowindex)* a->value;
     a=a->nextrow;
   }
   result(0,N)=outrate/(outrate-chain(N,N));
   factor=-chain(N,N)/(outrate-chain(N,N));
   for (int j=rows(x)-1; j>N;j--)
     result(0,j)=result(0,j)*factor;
  } 

//  continous_end ();
  return result;
}

REALMATRIX  equilib (const SPARSEREALMATRIX& x) {
  SPARSEREALMATRIX chain=x;
  REALMATRIX result;
  equilib_del(chain,result);
  return result;
}

void equilib_del (SPARSEREALMATRIX& chain,REALMATRIX& result) {
//  continous_begin ();
rprintf("Teil 1\n");

  REAL outrate,factor,divisor;
  SPM_ELEMENT* a;
  SPM_ELEMENT* b;
  SPM_ELEMENT* c;

  result=zeroes(1,1);

  for (unsigned int N=0; N < chain.maxrow; N++) {
    outrate=0;
    a=chain.rowvector[N];
    while ( a != NULL ) {
      outrate=outrate+a->value;
      a=a->nextcol;
    }
    sparse_set(chain,N,N,chain(N,N)-outrate);
  }
  for (int N=chain.maxrow-1; N >= 0; N--) {
    sparse_set(chain,N,chain.maxrow-1,1);
  }

rprintf("\nTeil 2, Fuellung jetzt: %d\n",chain.filled);


  for (unsigned int i=0; i < chain.maxrow; i++) {
    c=sparse_get_ptr(chain,i,i);
    divisor=c->value;
    a=c->nextcol;
    int first;
    while ( a != NULL ) {
      factor=a->value/divisor;
      b=c->nextrow;
 first=0;
  chain.hotspot=b;
      while ( b != NULL ) {
//if (chain.hotspot != NULL) rprintf("[%d,%d] ",chain.hotspot->rowindex,chain.hotspot->colindex); else rprintf("[NULL] ");
        sparse_set(chain,b->rowindex,a->colindex,
                    chain(b->rowindex,a->colindex)
                   -factor* b->value);
        b=b->nextrow;
      }
      a=a->nextcol;
    }
  }

rprintf("\nTeil 3, Fuellung jetzt: %d\n",chain.filled);

  result=zeroes(1,chain.maxrow);
  result(0,chain.maxrow-1)=1;
  for (int i=chain.maxrow-1; i >= 0; i--) {
    result(0,i)=result(0,i)/chain(i,i);
    a=chain.rowvector[i];
    while ( a != NULL && a->colindex< i) {
      result(0,a->colindex)=result(0,a->colindex) - a->value * result(0,i);
      a=a->nextcol;
    }
  }

rprintf("Teil 4\n");
//  continous_end ();
}

