/*-----------------------------------------------------------------------*/
/* Templates for general limits, lipschitz limits..                     */
/*-----------------------------------------------------------------------*/
#include <math.h>
#include <stdlib.h>
#include <stdarg.h>

#include "REALcore.h"
#include "REALMATRIX.h"

#include "COMPLEX.h"
#include "REALlimits.h"

template <class ARGUMENT, class RESULT> 
RESULT limit (RESULT f(long prec,const ARGUMENT&),
                           const ARGUMENT& x)

{ RESULT lim,limnew;
  sizetype limnew_error,element_error;
  sizetype lim_error,x_error;
  long prec_inc_old=prec_inc;
// volatile long element=actual_prec+prec_inc2;
  volatile long element=actual_prec;
  volatile long actual_prec_old=actual_prec;
  int inlimit_old=inlimit;
  volatile int firsttime=2;
  ITERATION_STACK LOCAL_STACK;
  ITERATION_STACK * old_STACK= ACTUAL_STACK;
  LOCAL_STACK.prec_policy=ACTUAL_STACK->prec_policy;

  inlimit=1;
  prec_inc=long(prec_inc * prec_factor)+prec_inc1;
  actual_prec=starting_prec+prec_inc; 
  x.geterror(x_error);
  DEBUG1(2,"starting limit_gen1...\n");
  ACTUAL_STACK= &LOCAL_STACK;
  while (1) {
    if (setjmp (ACTUAL_STACK->actual_jmp_buf)) {
      if ( firsttime==0) {
      DEBUG1(2,"computation failed, using best success\n");
      break;
      } else 
      if ( firsttime==2) {
      DEBUG1(2,"computation failed, trying normal p-sequence\n");
      element=4+starting_prec;
      prec_inc=0;
      firsttime=1;
      } else {
      DEBUG1(1,"computation of limit_gen1 failed totally\n");
      ACTUAL_STACK=old_STACK;
      actual_prec=actual_prec_old;
      prec_inc=prec_inc_old;
      REITERATE(0);
      }}
   else { 
    DEBUG2(2,"trying to compute limit_gen1 with precicion 2^(%ld)...\n",element);
    limnew=f(element,x);
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if (firsttime ==2 ) if ( limnew_error.exponent > actual_prec_old -prec_inc_old/prec_factor &&  limnew_error.exponent > x_error.exponent -prec_inc_old/prec_factor) {
    DEBUG0(2,{fprintf(stderr,"computation not precise enough (%ld*2^%ld), trying normal p-sequence\n",
                   limnew_error.mantissa,limnew_error.exponent);});
       element=4+starting_prec;
       prec_inc=0;
       firsttime=1;
    }
    if ( firsttime != 0 || sizetype_less(limnew_error,lim_error) ) {
      lim=limnew;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %ld*2^(%ld)\n", 
               lim_error.mantissa, lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      }
    firsttime=0; 
    if (element<=actual_prec_old)break;
    prec_inc=long(prec_inc * prec_factor)+prec_inc1;
    element=starting_prec+prec_inc; 
    }
  }
  ACTUAL_STACK=old_STACK;
  actual_prec=actual_prec_old;
  prec_inc=prec_inc_old;
  lim.seterror(lim_error);
  inlimit=inlimit_old;
  DEBUG0(2,{fprintf(stderr,"end of limit_gen1 with error %ld*2^(%ld)\n",
                   lim_error.mantissa,lim_error.exponent);});
  return lim;
}

template <class ARGUMENT, class RESULT> 
RESULT limit (RESULT f(long prec,const ARGUMENT&,long param),
                           const ARGUMENT& x,long param)

{ RESULT lim,limnew;
  sizetype limnew_error,element_error;
  sizetype lim_error,x_error;
  long prec_inc_old=prec_inc;
//  volatile long element=4+starting_prec;
  volatile long element=actual_prec;
  volatile long actual_prec_old=actual_prec;
  int inlimit_old=inlimit;
  volatile int firsttime=2;
  ITERATION_STACK LOCAL_STACK;
  ITERATION_STACK * old_STACK= ACTUAL_STACK;
  LOCAL_STACK.prec_policy=ACTUAL_STACK->prec_policy;

  inlimit=1;
  prec_inc=long(prec_inc * prec_factor)+prec_inc1;
  actual_prec=starting_prec+prec_inc; 
  x.geterror(x_error);
  DEBUG1(2,"starting limit_gen1...\n");
  ACTUAL_STACK= &LOCAL_STACK;
  while (1) {
    if (setjmp (ACTUAL_STACK->actual_jmp_buf)) {
      if ( firsttime==0) {
      DEBUG1(2,"computation failed, using best success\n");
      break;
      } else 
      if ( firsttime==2) {
      DEBUG1(2,"computation failed, trying normal p-sequence\n");
      element=4+starting_prec;
      prec_inc=0;
      firsttime=1;
      } else {
      DEBUG1(1,"computation of limit_gen1 failed totally\n");
      ACTUAL_STACK=old_STACK;
      actual_prec=actual_prec_old;
      prec_inc=prec_inc_old;
      REITERATE(0);
      }}
   else { 
    DEBUG2(2,"trying to compute limit_gen1 with precicion 2^(%ld)...\n",element);
    limnew=f(element,x,param);
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if (firsttime ==2 ) if ( limnew_error.exponent > actual_prec_old -prec_inc_old/prec_factor 
                       &&  limnew_error.exponent > x_error.exponent -prec_inc_old/prec_factor) {
    DEBUG0(2,{fprintf(stderr,"computation not precise enough (%ld*2^%ld), trying normal p-sequence\n",
                   limnew_error.mantissa,limnew_error.exponent);});
       element=4+starting_prec;
       prec_inc=0;
       firsttime=1;
    }
    if ( firsttime  != 0 || sizetype_less(limnew_error,lim_error) ) {
      lim=limnew;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %ld*2^(%ld)\n", 
               lim_error.mantissa, lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      }
    firsttime=0; 
    if (element<=actual_prec_old)break;
    prec_inc=long(prec_inc * prec_factor)+prec_inc1;
    element=starting_prec+prec_inc; 
    }
  }
  ACTUAL_STACK=old_STACK;
  actual_prec=actual_prec_old;
  prec_inc=prec_inc_old;
  lim.seterror(lim_error);
  inlimit=inlimit_old;
  DEBUG0(2,{fprintf(stderr,"end of limit_gen1 with error %ld*2^(%ld)\n",
                   lim_error.mantissa,lim_error.exponent);});
  return lim;
}


REAL limit         (REAL f(long, const REAL&, const REAL&),
                           const REAL& x, 
                           const REAL& y)
{ REAL lim,limnew;
  sizetype limnew_error,element_error;
  sizetype lim_error,x_error,y_error;
  long prec_inc_old=prec_inc;
//  volatile long element=4+starting_prec;
  volatile long element=actual_prec;
  volatile long actual_prec_old=actual_prec;
  int inlimit_old=inlimit;
  volatile int firsttime=2;
  ITERATION_STACK LOCAL_STACK;
  ITERATION_STACK * old_STACK= ACTUAL_STACK;
  LOCAL_STACK.prec_policy=ACTUAL_STACK->prec_policy;

  inlimit=1;
  prec_inc=long(prec_inc * prec_factor)+prec_inc1;
  actual_prec=starting_prec+prec_inc; 
  x.geterror(x_error);
  y.geterror(y_error);
  DEBUG1(2,"starting limit_gen2 ...\n");
  ACTUAL_STACK= &LOCAL_STACK;
  while (1) {
    if (setjmp (ACTUAL_STACK->actual_jmp_buf)) {
      if ( firsttime==0) {
      DEBUG1(2,"computation failed, using best success\n");
      break;
      } else 
      if ( firsttime==2) {
      DEBUG1(2,"computation failed, trying normal p-sequence\n");
      element=4+starting_prec;
      prec_inc=0;
      firsttime=1;
      } else {
      DEBUG1(1,"computation of limit_gen1 failed totally\n");
      ACTUAL_STACK=old_STACK;
      actual_prec=actual_prec_old;
      prec_inc=prec_inc_old;
      REITERATE(0);
      }}
   else { 
    DEBUG2(2,"trying to compute limit_gen2 with precision %ld...\n",element);
    limnew=f(element,x,y);
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if (firsttime ==2 ) if ( limnew_error.exponent > actual_prec_old -prec_inc_old/prec_factor 
     &&  limnew_error.exponent > x_error.exponent -prec_inc_old/prec_factor
     &&  limnew_error.exponent > y_error.exponent -prec_inc_old/prec_factor) {
    DEBUG0(2,{fprintf(stderr,"computation not precise enough (%ld*2^%ld), trying normal p-sequence\n",
                   limnew_error.mantissa,limnew_error.exponent);});
       element=4+starting_prec;
       prec_inc=0;
       firsttime=1;
    }
    if ( firsttime != 0 || sizetype_less(limnew_error,lim_error) ){
      lim=limnew;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %ld*2^(%ld)\n",
                   lim_error.mantissa,lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      }
    firsttime=0; 
    if (element<=actual_prec_old)break;
    prec_inc=long(prec_inc * prec_factor)+prec_inc1;
    element=starting_prec+prec_inc; 
    }
  }
  ACTUAL_STACK=old_STACK;
  actual_prec=actual_prec_old;
  prec_inc=prec_inc_old;
  lim.seterror(lim_error);
  inlimit=inlimit_old;
  DEBUG0(2,{fprintf(stderr,"end of limit_gen2 with error %ld*2^(%ld)\n",
                   lim_error.mantissa,lim_error.exponent);});
  return lim;
}


REAL limit (REAL f(long))
{ REAL lim;
  sizetype lim_error;
  volatile long l;
  volatile long actual_prec_old=actual_prec;
  volatile long prec_inc_old=prec_inc;
  volatile int inlimit_old=inlimit;
  ITERATION_STACK LOCAL_STACK;
  ITERATION_STACK * old_STACK= ACTUAL_STACK;
  LOCAL_STACK.prec_policy=ACTUAL_STACK->prec_policy;

  inlimit=1;
  DEBUG1(2,"starting limit_0 ...\n");
  l=0;
  ACTUAL_STACK= &LOCAL_STACK;
  prec_inc=prec_inc+prec_inc2;
  actual_prec=starting_prec+prec_inc; 
  while (1) {
    if (setjmp (ACTUAL_STACK->actual_jmp_buf)) {
      prec_inc=long(prec_inc * prec_factor)+prec_inc1;
      actual_prec=starting_prec+prec_inc; 
      DEBUG2(2,"limit_0 failed, increasing precision locally to %ld...\n",actual_prec);
   } else { 
    DEBUG2(2,"trying to compute limit_0 with precision %ld...\n",actual_prec);
   } 
   lim=f(actual_prec_old);
   DEBUG3(2,"getting result with local error %ld*2^(%ld)\n", lim.error.mantissa, lim.error.exponent);
   break;
  }
  ACTUAL_STACK=old_STACK;
  actual_prec=actual_prec_old;
  prec_inc=prec_inc_old;
  sizetype_set(lim_error,1,actual_prec_old);
  lim.adderror(lim_error);
  inlimit=inlimit_old;
  DEBUG0(2,{lim.geterror(lim_error);
            fprintf(stderr,"end of limit_0 with error %ld*2^(%ld)\n",
              lim_error.mantissa,lim_error.exponent);});
  return lim;
}

REAL limit_lip (REAL f(long,const REAL&),
            const long lip,
            const REAL& x)
{ REAL x_new,lim;
  sizetype lim_error;
  volatile long actual_prec_old=actual_prec;
  volatile long prec_inc_old=prec_inc;
  volatile int inlimit_old=inlimit;
  ITERATION_STACK LOCAL_STACK;
  ITERATION_STACK * old_STACK= ACTUAL_STACK;
  LOCAL_STACK.prec_policy=ACTUAL_STACK->prec_policy;

  x_new=x; 
  sizetype_exact(x_new.error);
  inlimit=1;
  DEBUG1(2,"starting limit_lip1 ...\n");
  ACTUAL_STACK= &LOCAL_STACK;
  prec_inc=prec_inc+prec_inc2;
  actual_prec=starting_prec+prec_inc; 
  while (1) {
    if (setjmp (ACTUAL_STACK->actual_jmp_buf)) {
      prec_inc=long(prec_inc * prec_factor)+prec_inc1;
      actual_prec=starting_prec+prec_inc; 
      DEBUG2(2,"limit_lip1 failed, increasing precision locally to %ld...\n",actual_prec);
    } else { 
      DEBUG2(2,"trying to compute limit_lip1 with precision %ld...\n",actual_prec);
    }
    lim=f(actual_prec_old,x_new);
    DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lim.error.mantissa, lim.error.exponent);
    break;
  }
  ACTUAL_STACK=old_STACK;
  actual_prec=actual_prec_old;
  prec_inc=prec_inc_old;
  sizetype_set(lim_error,1,actual_prec_old);
  lim.adderror(lim_error);
  sizetype_shift(lim_error,x.error,lip);
  lim.adderror(lim_error);
  inlimit=inlimit_old;
  DEBUG0(2,{lim.geterror(lim_error);
            fprintf(stderr,"end of limit_lip1 with error %ld*2^(%ld)\n",
              lim_error.mantissa,lim_error.exponent);
            x.geterror(lim_error);
            fprintf(stderr,"  error of argument: %ld*2^(%ld)\n",
              lim_error.mantissa,lim_error.exponent);});
  return lim;
}

REAL limit_lip     (REAL f(long, const REAL&, const REAL&),
                    long lip,
                    const REAL& x,
                    const REAL& y)
{ REAL x_new,y_new,lim;
  sizetype lim_error;
  volatile long actual_prec_old=actual_prec;
  volatile long prec_inc_old=prec_inc;
  volatile int inlimit_old=inlimit;
  ITERATION_STACK LOCAL_STACK;
  ITERATION_STACK * old_STACK= ACTUAL_STACK;
  LOCAL_STACK.prec_policy=ACTUAL_STACK->prec_policy;

  x_new=x;   sizetype_exact(x_new.error);
  y_new=y;   sizetype_exact(y_new.error);
  inlimit=1;
  DEBUG1(2,"starting limit_lip2 ...\n");
  ACTUAL_STACK= &LOCAL_STACK;
  prec_inc=prec_inc+prec_inc2;
  while (1) {
    if (setjmp (ACTUAL_STACK->actual_jmp_buf)) {
      prec_inc=long(prec_inc * prec_factor)+prec_inc1;
      actual_prec=starting_prec+prec_inc; 
      DEBUG2(2,"limit_lip2 failed, increasing precision locally to %ld...\n",actual_prec);
    } else { 
      DEBUG2(2,"trying to compute limit_lip2 with precision %ld...\n",actual_prec);
    }
    lim=f(actual_prec_old,x_new,y_new);
    DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lim.error.mantissa, lim.error.exponent);
    break;
  }
  ACTUAL_STACK=old_STACK;
  actual_prec=actual_prec_old;
  prec_inc=prec_inc_old;
  sizetype_set(lim_error,1,actual_prec_old);
  lim.adderror(lim_error);
  sizetype_shift(lim_error,x.error,lip);
  lim.adderror(lim_error);
  sizetype_shift(lim_error,y.error,lip);
  lim.adderror(lim_error);
  inlimit=inlimit_old;
  DEBUG0(2,{lim.geterror(lim_error);
            fprintf(stderr,"end of limit_lip1 with error %ld*2^(%ld)\n",
              lim_error.mantissa,lim_error.exponent);
            x.geterror(lim_error);
            fprintf(stderr,"  error of argument 1: %ld*2^(%ld)\n",
              lim_error.mantissa,lim_error.exponent);
            y.geterror(lim_error);
            fprintf(stderr,"  error of argument 2: %ld*2^(%ld)\n",
              lim_error.mantissa,lim_error.exponent);});
  return lim;
}

REAL lipschitz (REAL f(const REAL&),
            long lip,
            const REAL& x)
{ REAL x_new,lip_result;
  sizetype lip_error;

  x_new=x; 
  sizetype_exact(x_new.error);
  continous_begin ();
  DEBUG1(2,"starting lipschitz1 ...\n");
  lip_result=f(x_new);
  DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lip_result.error.mantissa, lip_result.error.exponent);
  continous_end ();
  sizetype_shift(lip_error,x.error,lip);
  lip_result.adderror(lip_error);
  DEBUG0(2,{lip_result.geterror(lip_error);
            fprintf(stderr,"end of lipschitz_1 with error %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);
            x.geterror(lip_error);
            fprintf(stderr,"  for argument with error %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);});
  return lip_result;
}

REAL lipschitz (REAL f(long, const REAL&),
            long lip,
            long k,
            const REAL& x)
{ REAL x_new,lip_result;
  sizetype lip_error;

  x_new=x; 
  sizetype_exact(x_new.error);
  continous_begin ();
  DEBUG1(2,"starting lipschitz1 ...\n");
  lip_result=f(k,x_new);
  DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lip_result.error.mantissa, lip_result.error.exponent);
  continous_end ();
  sizetype_shift(lip_error,x.error,lip);
  lip_result.adderror(lip_error);
  DEBUG0(2,{lip_result.geterror(lip_error);
            fprintf(stderr,"end of lipschitz_1 with error %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);
            x.geterror(lip_error);
            fprintf(stderr,"  error of argument: %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);});
  return lip_result;
}

REAL lipschitz (REAL f(const REAL&, const REAL&),
            long lip,
            const REAL& x,
            const REAL& y)
{ REAL x_new,y_new,lip_result;
  sizetype lip_error;

  x_new=x; 
  sizetype_exact(x_new.error);
  y_new=y;
  sizetype_exact(y_new.error);
  continous_begin ();
  DEBUG1(2,"starting lipschitz2 ...\n");
  lip_result=f(x_new,y_new);
  DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lip_result.error.mantissa, lip_result.error.exponent);
  continous_end ();
  sizetype_shift(lip_error,x.error,lip);
  lip_result.adderror(lip_error);
  sizetype_shift(lip_error,y.error,lip);
  lip_result.adderror(lip_error);
  DEBUG0(2,{lip_result.geterror(lip_error);
            fprintf(stderr,"end of lipschitz_2 with error %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);
            x.geterror(lip_error);
            fprintf(stderr,"  error of argument x: %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);
            y.geterror(lip_error);
            fprintf(stderr,"  error of argument y: %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);});
  return lip_result;
}

REAL lipschitz (REAL f(long, const REAL&, const REAL&),
            long lip,
            long k,
            const REAL& x,
            const REAL& y)
{ REAL x_new,y_new,lip_result;
  sizetype lip_error;

  x_new=x; 
  sizetype_exact(x_new.error);
  y_new=y;
  sizetype_exact(y_new.error);
  continous_begin ();
  DEBUG1(2,"starting lipschitz2 ...\n");
  lip_result=f(k,x_new,y_new);
  DEBUG3(2,"getting result with local error %ld*2^(%ld)\n",
             lip_result.error.mantissa, lip_result.error.exponent);
  continous_end ();
  sizetype_shift(lip_error,x.error,lip);
  lip_result.adderror(lip_error);
  sizetype_shift(lip_error,y.error,lip);
  lip_result.adderror(lip_error);
  DEBUG0(2,{lip_result.geterror(lip_error);
            fprintf(stderr,"end of lipschitz_2 with error %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);
            x.geterror(lip_error);
            fprintf(stderr,"  error of argument x: %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);
            y.geterror(lip_error);
            fprintf(stderr,"  error of argument y: %ld*2^(%ld)\n",
              lip_error.mantissa,lip_error.exponent);});
  return lip_result;
}

REAL limit_hint    (REAL f(long, const REAL&),
                    long hint,
                    const REAL& x)
{ REAL lim,limnew;
  sizetype limnew_error,element_error;
  sizetype lim_error;
  volatile long hintcopy=hint;
  volatile long element;
  volatile int inlimit_old=inlimit;
  volatile int success=0;
  ITERATION_STACK LOCAL_STACK;
  ITERATION_STACK * old_STACK= ACTUAL_STACK;
  LOCAL_STACK.prec_policy=ACTUAL_STACK->prec_policy;

  inlimit=1;
  DEBUG1(2,"starting limit_hint1...\n");
  x.geterror(element_error);
  element=max(element_error.exponent,actual_prec);
  ACTUAL_STACK= &LOCAL_STACK;
  while (1) {
    if (setjmp (ACTUAL_STACK->actual_jmp_buf)) {
    if (element>starting_prec)break;
    element=element+hintcopy; 
    }
    DEBUG2(2,"trying to compute limit_hint1 with precicion 2^(%ld)...\n",element);
    limnew=f(element,x);
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if ( (! success) || sizetype_less(limnew_error,lim_error) ) {
      lim=limnew;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %ld*2^(%ld)\n", 
               lim_error.mantissa, lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      hintcopy=2*hintcopy;
      }
    success=1; 
    if (element>starting_prec)break;
    element=element+hint; 
  }
  ACTUAL_STACK=old_STACK;
  if ( ! success) {
    DEBUG1(1,"computation of limit_hint1 failed\n");
    REITERATE(0);
  }
  lim.seterror(lim_error);
  inlimit=inlimit_old;
  DEBUG0(2,{fprintf(stderr,"end of limit_hint1 with error %ld*2^(%ld)\n",
                   lim_error.mantissa,lim_error.exponent);});
  return lim;
}


REAL limit_hint    (REAL f(long, const REAL&, const REAL&),
                    long hint,
                    const REAL& x, const REAL&y)
{ REAL lim,limnew;
  sizetype limnew_error,element_error;
  sizetype lim_error;
  volatile long hintcopy=hint;
  volatile long element;
  volatile int inlimit_old=inlimit;
  volatile int success=0;
  ITERATION_STACK LOCAL_STACK;
  ITERATION_STACK * old_STACK= ACTUAL_STACK;
  LOCAL_STACK.prec_policy=ACTUAL_STACK->prec_policy;

  inlimit=1;
  DEBUG1(2,"starting limit_hint1...\n");
  x.geterror(element_error);
  element=max(element_error.exponent,actual_prec);
  ACTUAL_STACK= &LOCAL_STACK;
  while (1) {
    if (setjmp (ACTUAL_STACK->actual_jmp_buf)) {
    if (element>starting_prec)break;
    element=element+hintcopy; 
    }
    DEBUG2(2,"trying to compute limit_hint1 with precicion 2^(%ld)...\n",element);
    limnew=f(element,x,y);
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if ( (! success) || sizetype_less(limnew_error,lim_error) ) {
      lim=limnew;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %ld*2^(%ld)\n", 
               lim_error.mantissa, lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      hintcopy=2*hintcopy;
      }
    success=1; 
    if (element>starting_prec)break;
    element=element+hint; 
  }
  ACTUAL_STACK=old_STACK;
  if ( ! success) {
    DEBUG1(1,"computation of limit_hint1 failed\n");
    REITERATE(0);
  }
  lim.seterror(lim_error);
  inlimit=inlimit_old;
  DEBUG0(2,{fprintf(stderr,"end of limit_hint1 with error %ld*2^(%ld)\n",
                   lim_error.mantissa,lim_error.exponent);});
  return lim;
}


REALMATRIX limit_lip (REALMATRIX f(long,const REALMATRIX&),
            long lip,
            const REALMATRIX& x)
{ REALMATRIX x_new,lim,limnew;
  sizetype lim_error;
  volatile long actual_prec_old=actual_prec;
  volatile long prec_inc_old=prec_inc;
  volatile int inlimit_old=inlimit;
  ITERATION_STACK LOCAL_STACK;
  ITERATION_STACK * old_STACK= ACTUAL_STACK;
  LOCAL_STACK.prec_policy=ACTUAL_STACK->prec_policy;

  sizetype_exact(lim_error);
  x_new=x; x_new.seterror(lim_error);
  inlimit=1;
  DEBUG1(2,"starting limit_matrix_lip1 ...\n");
  ACTUAL_STACK= &LOCAL_STACK;
  prec_inc=long(prec_inc * prec_factor)+prec_inc1;
  actual_prec=starting_prec+prec_inc; 
  prec_inc=prec_inc+prec_inc2;
  while (1) {
    if (setjmp (ACTUAL_STACK->actual_jmp_buf)) {
      prec_inc=long(prec_inc * prec_factor)+prec_inc1;
      actual_prec=starting_prec+prec_inc; 
      DEBUG2(2,"limit_lip1 failed, increasing precision locally to %ld...\n",actual_prec);
    } else { 
      DEBUG2(2,"trying to compute limit_matrix_lip1 with precision %ld...\n",actual_prec);
    }
    limnew=f(actual_prec_old,x_new);
    ACTUAL_STACK=old_STACK;
    lim=limnew;
    DEBUG0(2,{lim.geterror(lim_error);
              fprintf(stderr,"getting result with local error %ld*2^(%ld)\n",
                lim_error.mantissa,lim_error.exponent);});
    break;
  }
  actual_prec=actual_prec_old;
  prec_inc=prec_inc_old;
  sizetype_exact(lim_error);
  lim.adderror(lim_error);
  x.geterror(lim_error);
  sizetype_shift(lim_error,lim_error,lip);
  lim.adderror(lim_error);
  inlimit=inlimit_old;
  DEBUG0(2,{lim.geterror(lim_error);
            fprintf(stderr,"end of limit_matrix_lip1 with error %ld*2^(%ld)\n",
                lim_error.mantissa,lim_error.exponent);});

  return lim;
}



template <class ARGUMENT, class RESULT> 
RESULT limit_mv (RESULT f(long prec,
                                         long* choice,
                                         const ARGUMENT&),
                           const ARGUMENT& x)
{ RESULT lim,limnew;
  sizetype limnew_error,element_error;
  long choice=0;
  sizetype lim_error,x_error;
  volatile long prec_inc_old=prec_inc;
//  volatile long element=4+starting_prec;
  volatile long element=actual_prec;
  volatile long actual_prec_old=actual_prec;
  volatile int inlimit_old=inlimit;
  volatile int firsttime=2;
  ITERATION_STACK LOCAL_STACK;
  ITERATION_STACK * old_STACK= ACTUAL_STACK;
  LOCAL_STACK.prec_policy=ACTUAL_STACK->prec_policy;

  if ( ! inlimit ) {
	if (testanzahl > testaufrufe)
	{ 
		testaufrufe++;
		choice=testwert[testaufrufe];
	}
	else
	{
		testanzahl++;
		testaufrufe++;
		testwert[testaufrufe]=0;
	}
  }
  inlimit=1;
  prec_inc=long(prec_inc * prec_factor)+prec_inc1;
  actual_prec=starting_prec+prec_inc; 
  x.geterror(x_error);
  DEBUG1(2,"starting limit_mv...\n");
  ACTUAL_STACK= &LOCAL_STACK;
  while (1) {
    if (setjmp (ACTUAL_STACK->actual_jmp_buf)) {
      if ( firsttime==0) {
      DEBUG1(2,"computation failed, using best success\n");
      break;
      } else 
      if ( firsttime==2) {
      DEBUG1(2,"computation failed, trying normal p-sequence\n");
      element=4+starting_prec;
      prec_inc=0;
      firsttime=1;
      } else {
      DEBUG1(1,"computation of limit_gen1 failed totally\n");
      ACTUAL_STACK=old_STACK;
      actual_prec=actual_prec_old;
      prec_inc=prec_inc_old;
      REITERATE(0);
      }}
   else { 
    DEBUG2(2,"trying to compute limit_mv with precicion 2^(%ld)...\n",element);
    limnew=f(element,&choice,x);
    if ( ! inlimit_old) testwert[testaufrufe]=choice;
    sizetype_set(element_error,1,element);
    limnew.geterror(limnew_error);
    sizetype_inc(limnew_error,element_error);
    if (firsttime ==2 ) if ( limnew_error.exponent > actual_prec_old -prec_inc_old/prec_factor &&  limnew_error.exponent > x_error.exponent -prec_inc_old/prec_factor) {
    DEBUG0(2,{fprintf(stderr,"computation not precise enough (%ld*2^%ld), trying normal p-sequence\n",
                   limnew_error.mantissa,limnew_error.exponent);});
       element=4+starting_prec;
       prec_inc=0;
       firsttime=1;
    }
    if ( firsttime != 0 || sizetype_less(limnew_error,lim_error) ) {
      ACTUAL_STACK=old_STACK;
      lim=limnew;
      ACTUAL_STACK=&LOCAL_STACK;
      lim_error=limnew_error;
      DEBUG3(2,"getting result with error %ld*2^(%ld)\n", 
               lim_error.mantissa, lim_error.exponent);
      } else {
      DEBUG1(2,"computation successful, but no improvement\n");
      }
    firsttime=0; 
    if (element<=actual_prec_old)break;
    prec_inc=long(prec_inc * prec_factor)+prec_inc1;
    element=starting_prec+prec_inc; 
    }
  }
  ACTUAL_STACK= old_STACK;
  actual_prec=actual_prec_old;
  prec_inc=prec_inc_old;
  lim.seterror(lim_error);
  inlimit=inlimit_old;
  DEBUG0(2,{fprintf(stderr,"end of limit_mv with error %ld*2^(%ld)\n",
                   lim_error.mantissa,lim_error.exponent);});
  return lim;
}


// Instantiation of templates for multi-valued limits:

template COMPLEX limit_mv<COMPLEX,COMPLEX>
    (COMPLEX (*)(long, long *, COMPLEX const &), COMPLEX const &);

template REALMATRIX limit_mv<REALMATRIX,REALMATRIX>
    (REALMATRIX (*)(long, long *, REALMATRIX const &), REALMATRIX const &);

// Instantiation of templates for single-valued limits, one parameter:

template REALMATRIX limit <REALMATRIX,REALMATRIX>
    (REALMATRIX (*)(long, REALMATRIX const &), REALMATRIX const &);

template REAL limit <REAL,REAL>
    (REAL (*)(long, REAL const &), REAL const &);

template COMPLEX limit <COMPLEX,COMPLEX>
    (COMPLEX (*)(long, COMPLEX const &), COMPLEX const &);

// Instantiation of templates for single-valued limits, one parameter + long:

template REAL limit <REAL,REAL>
    (REAL (*)(long, REAL const &, long param), REAL const &,long param);
